# -*- coding: utf-8 -*-
import pathlib

import numpy as np
import yaml

here = pathlib.Path(__file__).parent
# Expose the default template
with open(here / "templatePDF.info", "r", encoding="utf-8") as o:
    template_info = yaml.safe_load(o)
with open(here / "Toy.info", "r", encoding="utf-8") as t:
    Toy_info = yaml.safe_load(t)


def load_info_from_file(pdfset_name):
    """
    Load the info file from a parent pdf.

    Parameters
    ----------
        pdfset_name : str
            parent pdf name

    Returns
    -------
        dict :
            info dictionary
    """
    import lhapdf  # pylint: disable=import-error, import-outside-toplevel

    src = pathlib.Path(lhapdf.paths()[0]) / pdfset_name
    with open(src / f"{pdfset_name}.info", "r", encoding="utf-8") as o:
        info = yaml.safe_load(o)
    return info


def load_blocks_from_file(pdfset_name, member):
    """
    Load a pdf from a parent pdf.

    Parameters
    ----------
        pdfset_name : str
            parent pdf name
        member : int
            pdf member

    Returns
    -------
        str :
            head of member file
        list(dict) :
            pdf blocks of data

    """
    import lhapdf  # pylint: disable=import-error, import-outside-toplevel

    pdf = lhapdf.mkPDF(pdfset_name, member)
    src = pathlib.Path(lhapdf.paths()[0]) / pdfset_name
    # read actual file
    cnt = []
    with open(
        src / f"{pdfset_name}_{pdf.memberID:04d}.dat", "r", encoding="utf-8"
    ) as o:
        cnt = o.readlines()
    # file head
    head = cnt[0]
    head_section = cnt.index("---\n")
    blocks = []
    while head_section < len(cnt) - 1:
        # section head
        next_head_section = cnt.index("---\n", head_section + 1)
        # determine participating pids
        xgrid = np.array(cnt[head_section + 1].strip().split(" "), dtype=np.float_)
        Qgrid = np.array(cnt[head_section + 2].strip().split(" "), dtype=np.float_)
        pids = np.array(cnt[head_section + 3].strip().split(" "), dtype=np.int_)
        # data
        data = []
        for line in cnt[head_section + 4 : next_head_section]:
            elems = np.fromstring(line.strip(), sep=" ")
            data.append(elems)
        Q2grid = [el * el for el in Qgrid]
        blocks.append(dict(xgrid=xgrid, Q2grid=Q2grid, pids=pids, data=np.array(data)))
        # iterate
        head_section = next_head_section
    return head, blocks



# -*- coding: utf-8 -*-
"""Weight 5 harmonic sums."""

import numba as nb

from . import f_functions as f
from .constants import log2, zeta2, zeta3, zeta4, zeta5
from .polygamma import cern_polygamma, symmetry_factor


@nb.njit(cache=True)
def S5(N):
    r"""Computes the harmonic sum :math:`S_5(N)`.

    .. math::
      S_5(N) = \sum\limits_{j=1}^N \frac 1 {j^5} = \frac 1 24 \psi_4(N+1)+\zeta(5)

    with :math:`\psi_4(N)` the 4th-polygamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    S_5 : complex
        Harmonic sum :math:`S_5(N)`

    See Also
    --------
    eko.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`

    """
    return zeta5 + 1.0 / 24.0 * cern_polygamma(N + 1.0, 4)


@nb.njit(cache=True)
def Sm5(N, hS5, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-5}(N)`.

    .. math::
      S_{-5}(N) = \sum\limits_{j=1}^N \frac {(-1)^j} {j^5}

    Parameters
    ----------
    N : complex
        Mellin moment
    hS5:  complex
        Harmonic sum :math:`S_{5}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm5 : complex
        Harmonic sum :math:`S_{-5}(N)`

    See Also
    --------
    eko.harmonic.w5.S5 : :math:`S_5(N)`

    """
    if is_singlet is None:
        return (
            1
            / 2**4
            * ((1 - (-1) ** N) / 2 * S5((N - 1) / 2) + ((-1) ** N + 1) / 2 * S5(N / 2))
            - hS5
        )
    if is_singlet:
        return 1 / 2**4 * S5(N / 2) - hS5
    return 1 / 2**4 * S5((N - 1) / 2) - hS5


@nb.njit(cache=True)
def S41(N, S1, S2, S3):
    r"""Analytic continuation of harmonic sum :math:`S_{4,1}(N)`.

    As implemented in eq 9.1 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S3: complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    S41 : complex
        Harmonic sum :math:`S_{4,1}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F9 :
        :math:`\mathcal{M}[(\text{Li}_4(x)/(x-1))_{+}](N)`

    """
    return -f.F9(N, S1) + S1 * zeta4 - S2 * zeta3 + S3 * zeta2


@nb.njit(cache=True)
def S311(N, S1, S2):
    r"""Analytic continuation of harmonic sum :math:`S_{3,1,1}(N)`.

    As implemented in eq 9.21 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
    S311 : complex
        Harmonic sum :math:`S_{3,1,1}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F11 :
        :math:`\mathcal{M}[(\text{S}_{2,2}(x)/(x-1))_{+}](N)`

    """
    return f.F11(N, S1, S2) + zeta3 * S2 - zeta4 / 4 * S1


@nb.njit(cache=True)
def S221(N, S1, S2, S21):
    r"""Analytic continuation of harmonic sum :math:`S_{2,2,1}(N)`.

    As implemented in eq 9.23 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S21: complex
        Harmonic sum :math:`S_{2,1}(N)`

    Returns
    -------
    S221 : complex
        Harmonic sum :math:`S_{2,2,1}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F11 :
        :math:`\mathcal{M}[(\text{S}_{2,2}(x)/(x-1))_{+}](N)`
    eko.harmonic.f_functions.F13 :
        :math:`\mathcal{M}[(\text{Li}_{2}^2(x)/(x-1))_{+}](N)`

    """
    return (
        -2 * f.F11(N, S1, S2)
        + 1 / 2 * f.F13(N, S1, S2)
        + zeta2 * S21
        - 3 / 10 * zeta2**2 * S1
    )


@nb.njit(cache=True)
def Sm221(N, S1, Sm1, S21, Sm21):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,2,1}(N)`.

    As implemented in eq 9.25 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    S21: complex
        Harmonic sum :math:`S_{2,1}(N)`
    Sm21: complex
        Harmonic sum :math:`S_{-2,1}(N)`

    Returns
    -------
    Sm221 : complex
        Harmonic sum :math:`S_{-2,2,1}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F14F12 :
        :math:`\mathcal{M}[(2 \text{S}_{2,2}(x) - 1/2 \text{Li}_{2}^2(x))/(x+1)](N)`

    """
    return (
        (-1) ** (N + 1) * (f.F14F12(N, S1, S21))
        + zeta2 * Sm21
        - 3 / 10 * zeta2**2 * Sm1
        - 0.119102
        + 0.0251709
    )


@nb.njit(cache=True)
def S21m2(N, S1, S2, Sm1, Sm2, Sm3, S21, Sm21, S2m1):
    r"""Analytic continuation of harmonic sum :math:`S_{2,1,-2}(N)`.

    As implemented in eq 9.26 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`
    Sm3: complex
        Harmonic sum :math:`S_{-3}(N)`
    S21: complex
        Harmonic sum :math:`S_{2,1}(N)`
    Sm21: complex
        Harmonic sum :math:`S_{-2,1}(N)`
    S2m1: complex
        Harmonic sum :math:`S_{2,-1}(N)`

    Returns
    -------
    S21m2 : complex
        Harmonic sum :math:`S_{2,1,-2}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F14F12 :
        :math:`\mathcal{M}[
            (\text{ln}(x) \text{S}_{1,2}(−x) − \text{Li}_2^2(−x)/2)/(x+1)
            ](N)`

    """
    return (
        (-1) ** (N) * f.F16(N, S1, Sm1, Sm2, Sm3, Sm21)
        - 1 / 2 * zeta2 * (S21 - S2m1)
        - (1 / 8 * zeta3 - 1 / 2 * zeta2 * log2) * (S2 - Sm2)
        + 1 / 8 * zeta2**2 * Sm1
        + 0.0854806
    )


@nb.njit(cache=True)
def S2111(N, S1, S2, S3):
    r"""Analytic continuation of harmonic sum :math:`S_{2,1,1,1}(N)`.

    As implemented in eq 9.33 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S3: complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    S2111 : complex
        Harmonic sum :math:`S_{2,1,1,1}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F17 :
        :math:`\mathcal{M}[(\text{S}_{1,3}(x)/(x-1))_{+}](N)`

    """
    return -f.F17(N, S1, S2, S3) + zeta4 * S1


@nb.njit(cache=True)
def Sm2111(N, S1, S2, S3, Sm1):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,1,1,1}(N)`.

    As implemented in eq 9.34 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S3: complex
        Harmonic sum :math:`S_{3}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`

    Returns
    -------
    Sm2111 : complex
        Harmonic sum :math:`S_{-2,1,1,1}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F18 :
        :math:`\mathcal{M}[\text{S}_{1,3}(x)/(x+1)](N)`

    """
    return (
        (-1) ** (N + 1) * f.F18(N, S1, S2, S3)
        + zeta4 * Sm1
        - 0.706186
        + 0.693147 * zeta4
    )


@nb.njit(cache=True)
def S23(N, S1, S2, S3):
    r"""Analytic continuation of harmonic sum :math:`S_{2,3}(N)`.

    As implemented in eq 9.3 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S1: complex
        Harmonic sum :math:`S_{1}(N)`
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    S3: complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    S23 : complex
        Harmonic sum :math:`S_{2,3}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F19 :
        :math:`\mathcal{M}[
            ((
                \text{ln}(x)[\text{S}_{1,2}(1-x) - \zeta_3]
                +3 [\text{S}_{1,3}(1-x) - \zeta_4]
            /(x-1))_{+}](N)`

    """
    return f.F19(N, S1, S2, S3) + 3 * zeta4 * S1


@nb.njit(cache=True)
def Sm23(N, Sm1, Sm2, Sm3, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-2,3}(N)`.

    As implemented in eq 9.4 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`
    Sm3: complex
        Harmonic sum :math:`S_{-3}(N)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N =
        1`), False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm23 : complex
        Harmonic sum :math:`S_{-2,3}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F19 :
        :math:`\mathcal{M}[
            (
                \text{ln}(x)[\text{S}_{1,2}(1-x) - \zeta_3]
                +3 [\text{S}_{1,3}(1-x) - \zeta_4]
            /(x+1)](N)`

    """
    eta = symmetry_factor(N, is_singlet)
    return (
        eta * f.F20(N, Sm1, Sm2, Sm3)
        + 3 * zeta4 * Sm1
        + 21 / 32 * zeta5
        + 3 * zeta4 * log2
        - 3 / 4 * zeta2 * zeta3
    )


@nb.njit(cache=True)
def S2m3(N, S2, Sm1, Sm2, Sm3):
    r"""Analytic continuation of harmonic sum :math:`S_{2,-3}(N)`.

    As implemented in eq 9.5 of :cite:`Blumlein:2009ta`.

    Parameters
    ----------
    N : complex
        Mellin moment
    S2: complex
        Harmonic sum :math:`S_{2}(N)`
    Sm1: complex
        Harmonic sum :math:`S_{-1}(N)`
    Sm2: complex
        Harmonic sum :math:`S_{-2}(N)`
    Sm3: complex
        Harmonic sum :math:`S_{-3}(N)`

    Returns
    -------
    S2m3 : complex
        Harmonic sum :math:`S_{2,-3}(N)`

    See Also
    --------
    eko.harmonic.f_functions.F19 :
        :math:`\mathcal{M}[
            ((
                1/2 \text{ln}^2(x) \text{Li}_2(-x)
                -2  \text{ln}(x) \text{Li}_3(-x)
                +3 \text{Li}_4(-x)
            )/(x-1)](N)`

    """
    return (
        (-1) ** (N + 1) * f.F21(N, Sm1, Sm2, Sm3)
        + 3 / 4 * zeta3 * (Sm2 - S2)
        - 21 / 8 * zeta4 * Sm1
        - 1.32056
    )



# -*- coding: utf-8 -*-
"""
This module defines the matching conditions for the |VFNS| evolution.
"""

from .. import basis_rotation as br
from .. import member


class MatchingCondition(member.OperatorBase):
    """
    Matching conditions for |PDF| at threshold.

    The continuation of the (formally) heavy non-singlet distributions
    with either the full singlet :math:`S` or the full valence :math:`V`
    is considered "trivial". Instead at |NNLO| additional terms ("non-trivial")
    enter.
    """

    @classmethod
    def split_ad_to_evol_map(
        cls,
        op_members,
        nf,
        q2_thr,
        intrinsic_range,
    ):
        """
        Create the instance from the |OME|.

        Parameters
        ----------
            op_members : eko.operator_matrix_element.OperatorMatrixElement.op_members
                Attribute of :class:`~eko.operator_matrix_element.OperatorMatrixElement`
                containing the |OME|
            nf : int
                number of active flavors *below* the threshold
            q2_thr: float
                threshold value
            intrinsic_range : list
                list of intrinsic quark pids
        """

        m = {
            "S.S": op_members[(100, 100)],
            "S.g": op_members[(100, 21)],
            "g.S": op_members[(21, 100)],
            "g.g": op_members[(21, 21)],
            "V.V": op_members[(200, 200)],
        }

        # add elements which are already active
        for f in range(2, nf + 1):
            n = f**2 - 1
            m[f"V{n}.V{n}"] = m["V.V"]
            m[f"T{n}.T{n}"] = m["V.V"]

        # activate the next heavy quark
        hq = br.quark_names[nf]
        m.update(
            {
                # f"{hq}-.V": op_members[(br.matching_hminus_pid, 200)],
                f"{hq}+.S": op_members[(br.matching_hplus_pid, 100)],
                f"{hq}+.g": op_members[(br.matching_hplus_pid, 21)],
            }
        )

        # intrinsic matching
        if len(intrinsic_range) != 0:
            op_id = member.OpMember.id_like(op_members[(200, 200)])
            for intr_fl in intrinsic_range:
                ihq = br.quark_names[intr_fl - 1]  # find name
                if intr_fl > nf + 1:
                    # keep the higher quarks as they are
                    m[f"{ihq}+.{ihq}+"] = op_id.copy()
                    m[f"{ihq}-.{ihq}-"] = op_id.copy()
                elif intr_fl == nf + 1:
                    # match the missing contribution from h+ and h-
                    m.update(
                        {
                            f"{ihq}+.{ihq}+": op_members[
                                (br.matching_hplus_pid, br.matching_hplus_pid)
                            ],
                            f"S.{ihq}+": op_members[(100, br.matching_hplus_pid)],
                            f"g.{ihq}+": op_members[(21, br.matching_hplus_pid)],
                            f"{ihq}-.{ihq}-": op_members[
                                (br.matching_hminus_pid, br.matching_hminus_pid)
                            ],
                            # f"V.{ihq}-": op_members[(200, br.matching_hminus_pid)],
                        }
                    )
        return cls.promote_names(m, q2_thr)



# -*- coding: utf-8 -*-
r"""
This module contains the |NLO| |OME| (OMEs)
for the matching conditions in the |VFNS|.
Heavy quark contribution for intrinsic evolution are taken from :cite:`Ball_2016`
and Mellin transformed with Mathematica.
The other matching conditions for the |VFNS| at :math:`\mu_F^2 \neq m_H^2`
are provided in :cite:`Buza_1998`.
"""
import numba as nb
import numpy as np

from ..constants import CF


@nb.njit(cache=True)
def A_hh(n, sx, L):
    r"""
    |NLO| heavy-heavy |OME| :math:`A_{HH}^{(1)}` defined as the
    mellin transform of :math:`K_{hh}` given in Eq. (20a) of :cite:`Ball_2016`.

    Parameters
    ----------
        n : complex
            Mellin moment
        sx : list
            harmonic sums cache
        L : float
            :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
        A_hh : complex
            |NLO| heavy-heavy |OME| :math:`A_{HH}^{(1)}`
    """
    S1m = sx[0][0] - 1 / n  # harmonics.S1(n - 1)
    S2m = sx[1][0] - 1 / n**2  # harmonics.S2(n - 1)
    ahh_l = (2 + n - 3 * n**2) / (n * (1 + n)) + 4 * S1m
    ahh = 2 * (
        2
        + 5 * n
        + n**2
        - 6 * n**3
        - 2 * n**4
        - 2 * n * (-1 - 2 * n + n**3) * S1m
    ) / (n * (1 + n)) ** 2 + 4 * (S1m**2 + S2m)
    return -CF * (ahh_l * L + ahh)


@nb.njit(cache=True)
def A_gh(n, L):
    r"""
    |NLO| gluon-heavy |OME| :math:`A_{gH}^{(1)}` defined as the
    mellin transform of :math:`K_{gh}` given in Eq. (20b) of :cite:`Ball_2016`.

    Parameters
    ----------
        n : complex
            Mellin moment
        L : float
            :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
        A_hg : complex
            |NLO| gluon-heavy |OME| :math:`A_{gH}^{(1)}`
    """

    agh_l1 = (2 + n + n**2) / (n * (n**2 - 1))
    agh_l0 = (-4 + 2 * n + n**2 * (15 + n * (3 + n - n**2))) / (
        n * (n**2 - 1)
    ) ** 2
    return 2 * CF * (agh_l0 + agh_l1 * L)


@nb.njit(cache=True)
def A_hg(n, L):
    r"""
    |NLO| heavy-gluon |OME| :math:`A_{Hg}^{S,(1)}` defined as the
    mellin transform of Eq. (B.2) from :cite:`Buza_1998`.

    Parameters
    ----------
        n : complex
            Mellin moment
        L : float
            :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
        A_hg : complex
            |NLO| heavy-gluon |OME| :math:`A_{Hg}^{S,(1)}`
    """
    den = 1.0 / (n * (n + 1) * (2 + n))
    num = 2 * (2 + n + n**2)
    return num * den * L


@nb.njit(cache=True)
def A_gg(L):
    r"""
    |NLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(1)}` defined as the
    mellin transform of Eq. (B.6) from :cite:`Buza_1998`.

    Parameters
    ----------
        L : float
            :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
        A_gg : complex
            |NLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(1)}`
    """
    return -2.0 / 3.0 * L


@nb.njit(cache=True)
def A_singlet(n, sx, L):
    r"""
      Computes the |NLO| singlet |OME|.

        .. math::
            A^{S,(1)} = \left(\begin{array}{cc}
            A_{gg,H}^{S,(1)} & 0  & A_{gH}^{(1)} \\
            0 & 0 & 0 \\
            A_{hg}^{S,(1)} & 0 & A_{HH}^{(1)}
            \end{array}\right)

      Parameters
      ----------
        n : complex
            Mellin moment
        sx : list
            harmonic sums cache containing: [[:math:`S_1`][:math:`S_2`]]
        L : float
            :math:`\ln(\mu_F^2 / m_h^2)`

      Returns
      -------
        A_S : numpy.ndarray
            |NLO| singlet |OME| :math:`A^{S,(1)}`

      See Also
      --------
        A_hg : :math:`A_{hg}^{S,(1)}`
        A_hh : :math:`A_{HH}^{(1)}`
        A_gg : :math:`A_{gg,H}^{S,(1)}`
        A_gh : :math:`A_{gH}^{(1)}`
    """
    A_S = np.array(
        [
            [A_gg(L), 0.0, A_gh(n, L)],
            [0 + 0j, 0 + 0j, 0 + 0j],
            [A_hg(n, L), 0.0, A_hh(n, sx, L)],
        ],
        np.complex_,
    )
    return A_S


@nb.njit(cache=True)
def A_ns(n, sx, L):
    r"""
      Computes the |NLO| non-singlet |OME| with intrinsic contributions.

      .. math::
          A^{NS,(1)} = \left(\begin{array}{cc}
            0 & 0 \\
            0 & A_{HH}^{(1)}
          \end{array}\right)

      Parameters
      ----------
        n : complex
            Mellin moment
        sx : list
            harmonic sums cache containing: [[:math:`S_1`][:math:`S_2`]]
        L : float
            :math:`\ln(\mu_F^2 / m_h^2)`
      Returns
      -------
        A_NS : numpy.ndarray
            |NLO| non-singlet |OME| :math:`A^{S,(1)}`

      See Also
      --------
        A_hh : :math:`A_{HH}^{(1)}`
    """
    return np.array([[0 + 0j, 0 + 0j], [0 + 0j, A_hh(n, sx, L)]], np.complex_)



# -*- coding: utf-8 -*-
"""
This module contains some harmonics sum.
Definitions are coming from :cite:`MuselliPhD,Bl_mlein_2000,Blumlein:2009ta`
"""
import numba as nb
import numpy as np

from .w1 import S1, Sm1
from .w2 import S2, Sm2
from .w3 import S3, S21, S2m1, Sm2m1, Sm3, Sm21
from .w4 import S4, S31, S211, Sm4, Sm22, Sm31, Sm211
from .w5 import S5, S23, S41, S221, S311, S2111, S2m3, S21m2, Sm5, Sm23, Sm221, Sm2111


@nb.njit(cache=True)
def base_harmonics_cache(n, is_singlet, max_weight=5, n_max_sums_weight=7):
    r"""
    Get the harmonics sums S basic cache.
    Only single index harmonics are computed and stored
    in the first (:math:`S_{n}`) or in the last column (:math:`S_{-n}`)

    Multi indices harmonics sums can be stored in the middle columns.

    Parameters
    ----------
        n : complex
            Mellin moment
        is_singlet: bool
            symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
            False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)
        max_weight : int
            max harmonics weight, max value 5 (default)
        n_max_sums_weight : int
            max number of harmonics sums for a given weight

    Returns
    -------
        h_cache : np.ndarray
            list of harmonics sums: (weights, n_max_sums_weight)
    """
    h_cache = np.zeros((max_weight, n_max_sums_weight), dtype=np.complex_)
    h_cache[:, 0] = sx(n, max_weight)
    if n_max_sums_weight > 1:
        h_cache[:, -1] = smx(n, h_cache[:, 0], is_singlet)
    return h_cache


@nb.njit(cache=True)
def sx(n, max_weight=5):
    """
    Get the harmonics sums S cache

    Parameters
    ----------
        n : complex
            Mellin moment
        max_weight : int
            max harmonics weight, max value 5 (default)

    Returns
    -------
        sx : np.ndarray
            list of harmonics sums (:math:`S_{1,..,w}`)
    """
    sx = np.zeros(max_weight, dtype=np.complex_)
    if max_weight >= 1:
        sx[0] = S1(n)
    if max_weight >= 2:
        sx[1] = S2(n)
    if max_weight >= 3:
        sx[2] = S3(n)
    if max_weight >= 4:
        sx[3] = S4(n)
    if max_weight >= 5:
        sx[4] = S5(n)
    return sx


@nb.njit(cache=True)
def smx(n, sx, is_singlet):
    r"""
    Get the harmonics S-minus cache

    Parameters
    ----------
        n : complex
            Mellin moment
        sx : numpy.ndarray
            List of harmonics sums: :math:`S_{1},\dots,S_{w}`
        is_singlet: bool
            symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
            False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)
    Returns
    -------
        smx : np.ndarray
            list of harmonics sums (:math:`S_{-1,..,-w}`)
    """
    max_weight = sx.size
    smx = np.zeros(max_weight, dtype=np.complex_)
    if max_weight >= 1:
        smx[0] = Sm1(n, sx[0], is_singlet)
    if max_weight >= 2:
        smx[1] = Sm2(n, sx[1], is_singlet)
    if max_weight >= 3:
        smx[2] = Sm3(n, sx[2], is_singlet)
    if max_weight >= 4:
        smx[3] = Sm4(n, sx[3], is_singlet)
    if max_weight >= 5:
        smx[4] = Sm5(n, sx[4], is_singlet)
    return smx


@nb.njit(cache=True)
def s3x(n, sx, smx, is_singlet):
    r"""
    Compute the weight 3 multi indices harmonics sums cache

    Parameters
    ----------
        n: complex
            Mellin moment
        sx : list
            List of harmonics sums: :math:`S_{1},S_{2}`
        smx : list
            List of harmonics sums: :math:`S_{-1},S_{-2}`
        is_singlet: bool
            symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
            False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
        s3x: np.ndarray
            list containing: :math:`S_{2,1},S_{2,-1},S_{-2,1},S_{-2,-1}`
    """
    return np.array(
        [
            S21(n, sx[0], sx[1]),
            S2m1(n, sx[1], smx[0], smx[1], is_singlet),
            Sm21(n, sx[0], smx[0], is_singlet),
            Sm2m1(n, sx[0], sx[1], smx[1]),
        ]
    )


@nb.njit(cache=True)
def s4x(n, sx, smx, is_singlet):
    r"""
    Compute the weight 4 multi indices harmonics sums cache

    Parameters
    ----------
        n: complex
            Mellin moment
        sx : list
            List of harmonics sums: :math:`S_{1},S_{2},S_{3},S_{4}`
        smx : list
            List of harmonics sums: :math:`S_{-1},S_{-2}`
        is_singlet: bool
            symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
            False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
        s4x: np.ndarray
            list containing: :math:`S_{3,1},S_{2,1,1},S_{-2,2},S_{-2,1,1},S_{-3,1}`
    """
    sm31 = Sm31(n, sx[0], smx[0], smx[1], is_singlet)
    return np.array(
        [
            S31(n, sx[0], sx[1], sx[2], sx[3]),
            S211(n, sx[0], sx[1], sx[2]),
            Sm22(n, sx[0], sx[1], smx[1], sm31, is_singlet),
            Sm211(n, sx[0], sx[1], smx[0], is_singlet),
            sm31,
        ]
    )


@nb.njit(cache=True)
def compute_cache(n, max_weight, is_singlet):
    r"""Get the harmonics sums cache.

    Parameters
    ----------
    n: complex
        Mellin moment
    max_weight: int
        maximum weight to compute [2,3,4,5]
    is_singlet: bool
        symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
        False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    list
        harmonic sums cache. At |N3LO| it contains:

        .. math ::
            [[S_1,S_{-1}],
            [S_2,S_{-2}],
            [S_{3}, S_{2,1}, S_{2,-1}, S_{-2,1}, S_{-2,-1}, S_{-3}],
            [S_{4}, S_{3,1}, S_{2,1,1}, S_{-2,-2}, S_{-3, 1}, S_{-4}],]
            [S_{5}, S_{-5}]

    """
    # max number of harmonics sum of a given weight for a given max weight.
    n_max_sums_weight = {2: 1, 3: 3, 4: 7, 5: 7}
    sx = base_harmonics_cache(n, is_singlet, max_weight, n_max_sums_weight[max_weight])
    if max_weight == 3:
        # Add Sm21 to cache
        sx[2, 1] = Sm21(n, sx[0, 0], sx[0, -1], is_singlet)
    elif max_weight >= 4:
        # Add weight 3 and 4 to cache
        sx[2, 1:-2] = s3x(n, sx[:, 0], sx[:, -1], is_singlet)
        sx[3, 1:-1] = s4x(n, sx[:, 0], sx[:, -1], is_singlet)
    # return list of list keeping the non zero values
    return [[el for el in sx_list if el != 0] for sx_list in sx]



# -*- coding: utf-8 -*-
"""
This module contains the definitions of the
:doc:`Flavor Basis and Evolution Basis </theory/FlavorSpace>`.
"""

import numpy as np

flavor_basis_pids = tuple([22] + list(range(-6, -1 + 1)) + [21] + list(range(1, 6 + 1)))
r"""
Sorted elements in Flavor Basis as |pid|.

Definition: `here <https://pdg.lbl.gov/2019/reviews/rpp2019-rev-monte-carlo-numbering.pdf>`_

corresponding |PDF| : :math:`\gamma, \bar t, \bar b, \bar c, \bar s, \bar u, \bar d, g,
d, u, s, c, b, t`
"""

flavor_basis_names = (
    "ph",
    "tbar",
    "bbar",
    "cbar",
    "sbar",
    "ubar",
    "dbar",
    "g",
    "d",
    "u",
    "s",
    "c",
    "b",
    "t",
)
"""String representation of :data:`flavor_basis_pids`."""


quark_names = "".join(flavor_basis_names[-6:])

evol_basis = (
    "ph",
    "S",
    "g",
    "V",
    "V3",
    "V8",
    "V15",
    "V24",
    "V35",
    "T3",
    "T8",
    "T15",
    "T24",
    "T35",
)
r"""
Sorted elements in Evolution Basis as :obj:`str`.

Definition: :ref:`here <theory/FlavorSpace:flavor basis>`.

corresponding |PDF| : :math:`\gamma, \Sigma, g, V, V_{3}, V_{8}, V_{15}, V_{24},
V_{35}, T_{3}, T_{8}, T_{15}, T_{24}, T_{35}`
"""

evol_basis_pids = tuple(
    [22, 100, 21, 200]
    + [200 + n**2 - 1 for n in range(2, 6 + 1)]
    + [100 + n**2 - 1 for n in range(2, 6 + 1)]
)
"""|pid| representation of :data:`evol_basis`."""

non_singlet_pids_map = {"ns-": 10201, "ns+": 10101, "nsV": 10200}

singlet_labels = ((100, 100), (100, 21), (21, 100), (21, 21))
non_singlet_labels = (
    (non_singlet_pids_map["ns-"], 0),
    (non_singlet_pids_map["ns+"], 0),
    (non_singlet_pids_map["nsV"], 0),
)
full_labels = (*singlet_labels, *non_singlet_labels)
anomalous_dimensions_basis = full_labels
r"""
Sorted elements in Anomalous Dimensions Basis as :obj:`str`.
"""
matching_hplus_pid = 90
matching_hminus_pid = 91

# Tranformation from physical basis to QCD evolution basis
rotate_flavor_to_evolution = np.array(
    [
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -1, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, -0, -1, 1, 0, -1, 1, 0, -0, 0, 0],
        [0, 0, 0, 0, 2, -1, -1, 0, 1, 1, -2, -0, 0, 0],
        [0, 0, 0, 3, -1, -1, -1, 0, 1, 1, 1, -3, 0, 0],
        [0, 0, 4, -1, -1, -1, -1, 0, 1, 1, 1, 1, -4, 0],
        [0, 5, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, -5],
        [0, 0, 0, 0, 0, 1, -1, 0, -1, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, -2, 1, 1, 0, 1, 1, -2, 0, 0, 0],
        [0, 0, 0, -3, 1, 1, 1, 0, 1, 1, 1, -3, 0, 0],
        [0, 0, -4, 1, 1, 1, 1, 0, 1, 1, 1, 1, -4, 0],
        [0, -5, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, -5],
    ]
)
"""
Basis rotation matrix between :doc:`Flavor Basis and Evolution Basis </theory/FlavorSpace>`.
"""

map_ad_to_evolution = {
    (100, 100): ["S.S"],
    (100, 21): ["S.g"],
    (21, 100): ["g.S"],
    (21, 21): ["g.g"],
    (non_singlet_pids_map["nsV"], 0): ["V.V"],
    (non_singlet_pids_map["ns+"], 0): [
        "T3.T3",
        "T8.T8",
        "T15.T15",
        "T24.T24",
        "T35.T35",
    ],
    (non_singlet_pids_map["ns-"], 0): [
        "V3.V3",
        "V8.V8",
        "V15.V15",
        "V24.V24",
        "V35.V35",
    ],
}
"""
Map anomalous dimension sectors' names to their members
"""


def ad_projector(ad_lab, nf):
    """
    Build a projector (as a numpy array) for the given anomalous dimension
    sector.

    Parameters
    ----------
    ad_lab : str
        name of anomalous dimension sector
    nf : int
        number of light flavors

    Returns
    -------
    proj : np.ndarray
        projector over the specified sector
    """
    proj = np.zeros_like(rotate_flavor_to_evolution, dtype=float)
    l = map_ad_to_evolution[ad_lab]
    # restrict the evolution basis to light flavors
    # NOTE: the cut is only needed for "NS_p" and "NS_m", but the other lists
    # are 1-long so they are unaffected
    l = l[: (nf - 1)]

    for el in l:
        out_name, in_name = el.split(".")
        out_idx = evol_basis.index(out_name)
        in_idx = evol_basis.index(in_name)
        out = rotate_flavor_to_evolution[out_idx].copy()
        in_ = rotate_flavor_to_evolution[in_idx].copy()
        out[: 1 + (6 - nf)] = out[len(out) - (6 - nf) :] = 0
        in_[: 1 + (6 - nf)] = in_[len(in_) - (6 - nf) :] = 0
        proj += (out[:, np.newaxis] * in_) / (out @ out)

    return proj


def ad_projectors(nf):
    """
    Build projectors tensor (as a numpy array), collecting all the individual
    sector projectors.

    Parameters
    ----------
    nf : int
        number of light flavors

    Returns
    -------
    projs : np.ndarray
        projectors tensor
    """
    projs = []
    for ad in anomalous_dimensions_basis:
        projs.append(ad_projector(ad, nf))

    return np.array(projs)


def intrinsic_unified_evol_labels(nf):
    """
    Collect all labels in the intrinsic unified evolution basis.

    Parameters
    ----------
    nf : int
        number of light flavors

    Returns
    -------
    labels : list(str)
        active distributions
    """
    labels = [
        "ph",
        "g",
        "S",
        "V",
        "Sdelta",
        "Vdelta",
        "Td3",
        "Vd3",
    ]
    news = ["u3", "d8", "u8"]
    for j in range(nf, 6):
        q = quark_names[j]
        labels.extend([f"{q}+", f"{q}-"])
    for k in range(3, nf):
        new = news[k - 3]
        labels.extend([f"T{new}", f"V{new}"])

    return labels



# -*- coding: utf-8 -*-
r"""
This module contains the central operator classes.

See :doc:`Operator overview </code/Operators>`.
"""

import functools
import logging
import os
import time
from multiprocessing import Pool

import numba as nb
import numpy as np
from scipy import integrate

from .. import anomalous_dimensions as ad
from .. import basis_rotation as br
from .. import interpolation, mellin
from .. import scale_variations as sv
from ..kernels import non_singlet as ns
from ..kernels import singlet as s
from ..member import OpMember

logger = logging.getLogger(__name__)


@nb.njit(cache=True)
def select_singlet_element(ker, mode0, mode1):
    """
    Select element of the singlet matrix

    Parameters
    ----------
        ker : numpy.ndarray
            singlet integration kernel
        mode0 : int
            id for first sector element
        mode1 : int
            id for second sector element
    Returns
    -------
        ker : complex
            singlet integration kernel element
    """

    k = 0 if mode0 == 100 else 1
    l = 0 if mode1 == 100 else 1
    return ker[k, l]


spec = [
    ("is_singlet", nb.boolean),
    ("is_log", nb.boolean),
    ("logx", nb.float64),
    ("u", nb.float64),
]


@nb.experimental.jitclass(spec)
class QuadKerBase:
    """
    Manage the common part of Mellin inversion integral

    Parameters
    ----------
        u : float
            quad argument
        is_log : boolean
            is a logarithmic interpolation
        logx : float
            Mellin inversion point
        mode0 : str
            first sector element
    """

    def __init__(self, u, is_log, logx, mode0):
        self.is_singlet = mode0 in [100, 21, 90]
        self.is_log = is_log
        self.u = u
        self.logx = logx

    @property
    def path(self):
        """Returns the associated instance of :class:`eko.mellin.Path`"""
        return mellin.Path(self.u, self.logx, self.is_singlet)

    @property
    def n(self):
        """Returns the Mellin moment N"""
        return self.path.n

    def integrand(
        self,
        areas,
    ):
        """
        Get transformation to Mellin space integral

        Parameters
        ----------
            areas : tuple
                basis function configuration

        Returns
        -------
            base_integrand: complex
                common mellin inversion integrand
        """
        if self.logx == 0.0:
            return 0.0
        pj = interpolation.evaluate_grid(self.path.n, self.is_log, self.logx, areas)
        if pj == 0.0:
            return 0.0
        return self.path.prefactor * pj * self.path.jac


@nb.njit(cache=True)
def quad_ker(
    u,
    order,
    mode0,
    mode1,
    method,
    is_log,
    logx,
    areas,
    as1,
    as0,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    u : float
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    is_log : boolean
        is a logarithmic interpolation
    logx : float
        Mellin inversion point
    areas : tuple
        basis function configuration
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an itermediate threshold operator?

    Returns
    -------
    float
        evaluated integration kernel
    """
    ker_base = QuadKerBase(u, is_log, logx, mode0)
    integrand = ker_base.integrand(areas)
    if integrand == 0.0:
        return 0.0

    # compute the actual evolution kernel
    if ker_base.is_singlet:
        gamma_singlet = ad.gamma_singlet(order, ker_base.n, nf)
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_singlet = sv.exponentiated.gamma_variation(
                gamma_singlet, order, nf, L
            )
        ker = s.dispatcher(
            order,
            method,
            gamma_singlet,
            as1,
            as0,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv.expanded.singlet_variation(gamma_singlet, as1, order, nf, L)
            ) @ np.ascontiguousarray(ker)
        ker = select_singlet_element(ker, mode0, mode1)
    else:
        gamma_ns = ad.gamma_ns(order, mode0, ker_base.n, nf)
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv.exponentiated.gamma_variation(gamma_ns, order, nf, L)
        ker = ns.dispatcher(
            order,
            method,
            gamma_ns,
            as1,
            as0,
            nf,
            ev_op_iterations,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = sv.expanded.non_singlet_variation(gamma_ns, as1, order, nf, L) * ker

    # recombine everything
    return np.real(ker * integrand)


class Operator(sv.ModeMixin):
    """Internal representation of a single EKO.

    The actual matrices are computed upon calling :meth:`compute`.

    Parameters
    ----------
    config : dict
        configuration
    managers : dict
        managers
    nf : int
        number of active flavors
    q2_from : float
        evolution source
    q2_to : float
        evolution target
    mellin_cut : float
        cut to the upper limit in the mellin inversion
    is_threshold : bool
        is this an itermediate threshold operator?
    """

    log_label = "Evolution"
    # complete list of possible evolution operators labels
    full_labels = br.full_labels

    def __init__(
        self, config, managers, nf, q2_from, q2_to, mellin_cut=5e-2, is_threshold=False
    ):
        self.config = config
        self.managers = managers
        self.nf = nf
        self.q2_from = q2_from
        self.q2_to = q2_to
        # TODO make 'cut' external parameter?
        self._mellin_cut = mellin_cut
        self.is_threshold = is_threshold
        self.op_members = {}
        self.order = config["order"]

    @property
    def n_pools(self):
        n_pools = self.config["n_integration_cores"]
        if n_pools > 0:
            return n_pools
        # so we subtract from the maximum number
        return max(os.cpu_count() + n_pools, 1)

    @property
    def fact_to_ren(self):
        r"""Returns the factor :math:`(\mu_F/\mu_R)^2`"""
        return self.config["fact_to_ren"]

    @property
    def int_disp(self):
        """Returns the interpolation dispatcher"""
        return self.managers["interpol_dispatcher"]

    @property
    def grid_size(self):
        """Returns the grid size"""
        return self.int_disp.xgrid.size

    def mur2_shift(self, q2):
        """Computes shifted renormalization scale.

        Parameters
        ----------
        q2 : float
            factorization scale

        Returns
        -------
        float
            renormalization scale
        """
        if self.sv_mode == sv.Modes.exponentiated:
            return q2 / self.fact_to_ren
        return q2

    @property
    def a_s(self):
        """Returns the computed values for :math:`a_s`"""
        sc = self.managers["strong_coupling"]
        a0 = sc.a_s(
            self.mur2_shift(self.q2_from), fact_scale=self.q2_from, nf_to=self.nf
        )
        a1 = sc.a_s(self.mur2_shift(self.q2_to), fact_scale=self.q2_to, nf_to=self.nf)
        return (a0, a1)

    @property
    def labels(self):
        """Compute necessary sector labels to compute.

        Returns
        -------
        list(str)
            sector labels
        """
        labels = []
        # the NS sector is dynamic
        if self.config["debug_skip_non_singlet"]:
            logger.warning("%s: skipping non-singlet sector", self.log_label)
        else:
            # add + as default
            labels.append(br.non_singlet_labels[1])
            if self.order[0] >= 2:  # - becomes different starting from NLO
                labels.append(br.non_singlet_labels[0])
            if self.order[0] >= 3:  # v also becomes different starting from NNLO
                labels.append(br.non_singlet_labels[2])
        # singlet sector is fixed
        if self.config["debug_skip_singlet"]:
            logger.warning("%s: skipping singlet sector", self.log_label)
        else:
            labels.extend(br.singlet_labels)
        return labels

    def quad_ker(self, label, logx, areas):
        """Partially initialized integrand function.

        Parameters
        ----------
        label: tuple
            operator element pids
        logx: float
            Mellin inversion point
        areas : tuple
            basis function configuration

        Returns
        -------
        functools.partial
            partially initialized integration kernel

        """
        return functools.partial(
            quad_ker,
            order=self.order,
            mode0=label[0],
            mode1=label[1],
            method=self.config["method"],
            is_log=self.int_disp.log,
            logx=logx,
            areas=areas,
            as1=self.a_s[1],
            as0=self.a_s[0],
            nf=self.nf,
            L=np.log(self.fact_to_ren),
            ev_op_iterations=self.config["ev_op_iterations"],
            ev_op_max_order=self.config["ev_op_max_order"],
            sv_mode=self.sv_mode,
            is_threshold=self.is_threshold,
        )

    def initialize_op_members(self):
        """Init all ops with identity or zeros if we skip them"""
        eye = OpMember(
            np.eye(self.grid_size), np.zeros((self.grid_size, self.grid_size))
        )
        zero = OpMember(*[np.zeros((self.grid_size, self.grid_size))] * 2)
        for n in self.full_labels:
            if n in self.labels:
                # non-singlet evolution and diagonal op are identities
                if n in br.non_singlet_labels or n[0] == n[1]:
                    self.op_members[n] = eye.copy()
                else:
                    self.op_members[n] = zero.copy()
            else:
                self.op_members[n] = zero.copy()

    def run_op_integration(
        self,
        log_grid,
    ):
        """Run the integration for each grid point

        Parameters
        ----------
        log_grid : tuple(k, logx)
            log grid point with relative index

        Returns
        -------
        list
            computed operators at the give grid point

        """
        column = []
        k, logx = log_grid
        start_time = time.perf_counter()
        # iterate basis functions
        for l, bf in enumerate(self.int_disp):
            if k == l and l == self.grid_size - 1:
                continue
            temp_dict = {}
            # iterate sectors
            for label in self.labels:
                res = integrate.quad(
                    self.quad_ker(label, logx, bf.areas_representation),
                    0.5,
                    1.0 - self._mellin_cut,
                    epsabs=1e-12,
                    epsrel=1e-5,
                    limit=100,
                    full_output=1,
                )
                temp_dict[label] = res[:2]
            column.append(temp_dict)

        logger.info(
            "%s: computing operators - %u/%u took: %6f s",
            self.log_label,
            k + 1,
            self.grid_size,
            (time.perf_counter() - start_time),
        )
        return column

    def compute(self):
        """Compute the actual operators (i.e. run the integrations)."""
        self.initialize_op_members()

        # skip computation ?
        if np.isclose(self.q2_from, self.q2_to):
            logger.info(
                "%s: skipping unity operator at %e", self.log_label, self.q2_from
            )
            self.copy_ns_ops()
            return

        logger.info(
            "%s: computing operators %e -> %e, nf=%d",
            self.log_label,
            self.q2_from,
            self.q2_to,
            self.nf,
        )
        logger.info(
            "%s: µ_R^2 distance: %e -> %e",
            self.log_label,
            self.mur2_shift(self.q2_from),
            self.mur2_shift(self.q2_to),
        )
        if self.sv_mode.name != "unvaried":
            logger.info(
                "Scale Variation: (µ_F/µ_R)^2 = %e, mode: %s",
                self.fact_to_ren,
                self.sv_mode.name,
            )
        logger.info(
            "%s: a_s distance: %e -> %e", self.log_label, self.a_s[0], self.a_s[1]
        )
        logger.info(
            "%s: order: (%d, %d), solution strategy: %s",
            self.log_label,
            self.order[0],
            self.order[1],
            self.config["method"],
        )

        self.integrate()
        # copy non-singlet kernels, if necessary
        self.copy_ns_ops()

    def integrate(
        self,
    ):
        """Run the integration"""
        tot_start_time = time.perf_counter()

        # run integration in parallel for each grid point
        # or avoid opening a single pool
        args = (self.run_op_integration, enumerate(np.log(self.int_disp.xgrid_raw)))
        if self.n_pools == 1:
            res = map(*args)
        else:
            with Pool(self.n_pools) as pool:
                res = pool.map(*args)

        # collect results
        for k, row in enumerate(res):
            for l, entry in enumerate(row):
                for label, (val, err) in entry.items():
                    self.op_members[label].value[k][l] = val
                    self.op_members[label].error[k][l] = err

        # closing comment
        logger.info(
            "%s: Total time %f s",
            self.log_label,
            time.perf_counter() - tot_start_time,
        )

    def copy_ns_ops(self):
        """Copy non-singlet kernels, if necessary"""
        if self.order[0] == 1:  # in LO +=-=v
            for label in ["nsV", "ns-"]:
                self.op_members[
                    (br.non_singlet_pids_map[label], 0)
                ].value = self.op_members[
                    (br.non_singlet_pids_map["ns+"], 0)
                ].value.copy()
                self.op_members[
                    (br.non_singlet_pids_map[label], 0)
                ].error = self.op_members[
                    (br.non_singlet_pids_map["ns+"], 0)
                ].error.copy()
        elif self.order[0] == 2:  # in NLO -=v
            self.op_members[
                (br.non_singlet_pids_map["nsV"], 0)
            ].value = self.op_members[(br.non_singlet_pids_map["ns-"], 0)].value.copy()
            self.op_members[
                (br.non_singlet_pids_map["nsV"], 0)
            ].error = self.op_members[(br.non_singlet_pids_map["ns-"], 0)].error.copy()



# -*- coding: utf-8 -*-
"""
Implementations of some mellin transformations :cite:`MuselliPhD`
appearing in the analytic continuation of harmonics sums of weight = 3,4.
"""
import numba as nb
import numpy as np

from . import w1
from .constants import log2, zeta2, zeta3
from .polygamma import recursive_harmonic_sum as s

a1 = np.array(
    [
        0.999999974532238,
        -0.499995525889840,
        0.333203435557262,
        -0.248529457782640,
        0.191451164719161,
        -0.137466222728331,
        0.0792107412244877,
        -0.0301109656912626,
        0.00538406208663153,
        0.0000001349586745,
    ]
)

c1 = np.array(
    [
        2.2012182965269744e-8,
        2.833327652357064,
        -1.8330909624101532,
        0.7181879191200942,
        -0.0280403220046588,
        -0.181869786537805,
        0.532318519269331,
        -1.07281686995035,
        1.38194913357518,
        -1.11100841298484,
        0.506649587198046,
        -0.100672390783659,
    ]
)

c3 = np.array(
    [
        0,
        1.423616247405256,
        -0.08001203559240111,
        -0.39875367195395994,
        0.339241791547134,
        -0.0522116678353452,
        -0.0648354706049337,
        0.0644165053822532,
        -0.0394927322542075,
        0.0100879370657869,
    ]
)

p11 = np.array([11.0 / 6.0, -3.0, 3.0 / 2.0, -1.0 / 3.0])
p32 = np.array([-25.0 / 24.0, 2.0, -3.0 / 2.0, 2.0 / 3.0, -1.0 / 8.0])
p31 = np.array([205.0 / 144.0, -25.0 / 12.0, 23.0 / 24.0, -13.0 / 36.0, 1.0 / 16])


@nb.njit(cache=True)
def mellin_g3(N, S1):
    r"""
    Computes the Mellin transform of :math:`\text{Li}_2(x)/(1+x)`.

    This function appears in the analytic continuation of the harmonic sum
    :math:`S_{-2,1}(N)` which in turn appears in the |NLO| anomalous dimension
    (see :ref:`theory/mellin:harmonic sums`).

    Parameters
    ----------
        N : complex
            Mellin moment
        S1: complex
            Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
        mellin_g3 : complex
            approximate Mellin transform :math:`\mathcal{M}[\text{Li}_2(x)/(1+x)](N)`

    Note
    ----
        We use the name from :cite:`MuselliPhD`, but not his implementation - rather we use the
        Pegasus :cite:`Vogt:2004ns` implementation.
    """
    cs = [1.0000e0, -0.9992e0, 0.9851e0, -0.9005e0, 0.6621e0, -0.3174e0, 0.0699e0]
    g3 = 0
    for j, c in enumerate(cs):
        Nj = N + j
        g3 += c * (zeta2 - s(S1, N, j, 1) / Nj) / Nj
    return g3


@nb.njit(cache=True)
def mellin_g4(N):
    r"""
    Computes the Mellin transform of :math:`\text{Li}_2(-x)/(1+x)`.

    Implementation and definition in B.5.25 of :cite:`MuselliPhD` or
    in eq 61 of :cite:`Bl_mlein_2000`, but none of them is fully correct.


    Parameters
    ----------
        N : complex
            Mellin moment

    Returns
    -------
        mellin_g4 : complex
            Mellin transform :math:`\mathcal{M}[\text{Li}_2(-x)/(1+x)](N)`
    """
    g4 = -1 / 2 * zeta2 * log2
    for k, ak in enumerate(a1):
        Nk = N + k + 1
        beta = 1 / 2 * (w1.S1((Nk) / 2) - w1.S1((Nk - 1) / 2))
        g4 += ak * (N / Nk * zeta2 / 2 + (k + 1) / Nk**2 * (log2 - beta))
    return g4


@nb.njit(cache=True)
def mellin_g5(N, S1, S2):
    r"""
    Computes the Mellin transform of :math:`(\text{Li}_2(x)ln(x))/(1+x)`.

    Implementation and definition in B.5.26 of :cite:`MuselliPhD` or
    in eq 62 of :cite:`Bl_mlein_2000`, but none of them is fully correct.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1: complex
            Harmonic sum :math:`S_{1}(N)`
        S2: complex
            Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
        mellin_g5 : complex
            Mellin transform :math:`\mathcal{M}[(\text{Li}_2(x)ln(x))/(1+x)](N)`
    """
    g5 = 0.0
    for k, ak in enumerate(a1):
        Nk = N + k + 1
        poly1nk = -s(S2, N, k + 1, 2) + zeta2
        g5 -= ak * ((k + 1) / Nk**2 * (zeta2 + poly1nk - 2 * s(S1, N, k + 1, 1) / Nk))
    return g5


@nb.njit(cache=True)
def mellin_g6(N, S1):
    r"""
    Computes the Mellin transform of :math:`\text{Li}_3(x)/(1+x)`.

    Implementation and definition in B.5.27 of :cite:`MuselliPhD` or
    in eq 63 of :cite:`Bl_mlein_2000`, but none of them is fully correct.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1: complex
            Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
        mellin_g6 : complex
            Mellin transform :math:`\mathcal{M}[\text{Li}_3(x)/(1+x)](N)`


    """
    g6 = zeta3 * log2
    for k, ak in enumerate(a1):
        Nk = N + k + 1
        g6 -= ak * (
            N / Nk * zeta3 + (k + 1) / Nk**2 * (zeta2 - s(S1, N, k + 1, 1) / Nk)
        )
    return g6


@nb.njit(cache=True)
def mellin_g8(N, S1, S2):
    r"""
    Computes the Mellin transform of :math:`S_{1,2}(x)/(1+x)`.

    Implementation and definition in B.5.29 of :cite:`MuselliPhD` or
    in eq 65 of :cite:`Bl_mlein_2000`, but none of them is fully correct.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1: complex
            Harmonic sum :math:`S_{1}(N)`
        S2: complex
            Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
        mellin_g8 : complex
            Mellin transform :math:`\mathcal{M}[S_{1,2}(x)/(1+x)](N)`
    """
    g8 = zeta3 * log2
    for k, ak in enumerate(a1):
        Nk = N + k + 1
        g8 -= ak * (
            N / Nk * zeta3
            + (k + 1) / Nk**2 * 1 / 2 * (s(S1, N, k + 1, 1) ** 2 + s(S2, N, k + 1, 2))
        )
    return g8


@nb.njit(cache=True)
def mellin_g18(N, S1, S2):
    r"""
    Computes the Mellin transform of :math:`-(\text{Li}_2(x) - \zeta_2)/(1-x)`.

    Implementation and definition in eq 124 of :cite:`Bl_mlein_2000`

    Note: comparing to :cite:`Bl_mlein_2000`, we believe :cite:`MuselliPhD`
    was not changing the notations of :math:`P^{(1)}_{2}` to :math:`P^{(1)}_{1}`.
    So we implement eq 124 of :cite:`Bl_mlein_2000` but using :cite:`MuselliPhD`
    notation.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1 : complex
            Harmonic sum :math:`S_{1}(N)`
        S2 : complex
            Harmonic sum :math:`S_{2}(N)`
    Returns
    -------
        mellin_g18 : complex
            Mellin transform :math:`\mathcal{M}[-(\text{Li}_2(x) - \zeta_2)/(1-x)](N)`
    """
    g18 = (S1**2 + S2) / (N) - zeta2 * S1
    for k, ck in enumerate(c1):
        Nk = N + k
        g18 += ck * (N) / (Nk) * s(S1, N, k, 1)
    for k, p11k in enumerate(p11):
        Nk = N + k
        g18 -= p11k * (N) / (Nk) * (s(S1, N, k, 1) ** 2 + s(S2, N, k, 2))
    return g18


@nb.njit(cache=True)
def mellin_g19(N, S1):
    r"""
    Computes the Mellin transform of :math:`-(\text{Li}_2(-x) + \zeta_2/2)/(1-x)`.

    Implementation and definition in B.5.40 of :cite:`MuselliPhD` or in 125 of
    :cite:`Bl_mlein_2000`, but none of them is fully correct.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1 : complex
           Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
        mellin_g19 : complex
            Mellin transform :math:`\mathcal{M}[-(\text{Li}_2(-x) + \zeta_2/2)/(1-x)](N)`
    """
    g19 = 1 / 2 * zeta2 * S1
    for k, ak in enumerate(a1):
        g19 -= ak / (k + 1) * s(S1, N, k + 1, 1)
    return g19


@nb.njit(cache=True)
def mellin_g21(N, S1, S2, S3):
    r"""
    Computes the Mellin transform of :math:`-(S_{1,2}(x) - \zeta_3)/(1-x)`.

    Implementation and definition in B.5.42 of :cite:`MuselliPhD`

    Note: comparing to :cite:`Bl_mlein_2000`, we believe :cite:`MuselliPhD`
    was not changing the notations of :math:`P^{(3)}_{2}` to :math:`P^{(3)}_{1}`
    and :math:`P^{(3)}_{3}` to :math:`P^{(3)}_{2}`.
    So we implement eq 127 of :cite:`Bl_mlein_2000` but using :cite:`MuselliPhD`
    notation.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1 : complex
            Harmonic sum :math:`S_{1}(N)`
        S2 : complex
            Harmonic sum :math:`S_{2}(N)`
        S3 : complex
            Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
        mellin_g21 : complex
            Mellin transform :math:`\mathcal{M}[-(S_{1,2}(x) - \zeta_3)/(1-x)](N)`
    """
    g21 = -zeta3 * S1 + (S1**3 + 3 * S1 * S2 + 2 * S3) / (2 * N)
    for k, ck in enumerate(c3):
        Nk = N + k
        g21 += ck * N / Nk * s(S1, N, k, 1)
    for k in range(0, 5):
        Nk = N + k
        S1nk = s(S1, N, k, 1)
        S2nk = s(S2, N, k, 2)
        S3nk = s(S3, N, k, 3)
        g21 += (
            N
            / Nk
            * (
                p32[k] * (S1nk**3 + 3 * S1nk * S2nk + 2 * S3nk)
                - p31[k] * (S1nk**2 + S2nk)
            )
        )
    return g21


@nb.njit(cache=True)
def mellin_g22(N, S1, S2, S3):
    r"""
    Computes the Mellin transform of :math:`-(\text{Li}_2(x) ln(x))/(1-x)`.

    Implementation and definition in B.5.43 of :cite:`MuselliPhD`

    Note: comparing to :cite:`Bl_mlein_2000`, we believe :cite:`MuselliPhD`
    was not changing the notations of :math:`P^{(1)}_{2}` to :math:`P^{(1)}_{1}`
    So we implement eq 128 of :cite:`Bl_mlein_2000` but using :cite:`MuselliPhD`
    notation.

    Parameters
    ----------
        N : complex
            Mellin moment
        S1 : complex
            Harmonic sum :math:`S_{1}(N)`
        S2 : complex
            Harmonic sum :math:`S_{2}(N)`
        S3 : complex
            Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
        mellin_g22 : complex
            Mellin transform :math:`\mathcal{M}[-(\text{Li}_2(x) ln(x))/(1-x)](N)`
    """
    g22 = 0.0
    for k, ck in enumerate(c1):
        poly1nk = -s(S2, N, k, 2) + zeta2
        g22 += ck * poly1nk
    for k, p11k in enumerate(p11):
        S1nk = s(S1, N, k, 1)
        poly1nk = -s(S2, N, k, 2) + zeta2
        poly2nk = 2 * s(S3, N, k, 3) + zeta3
        g22 -= p11k * (S1nk * poly1nk - 1 / 2 * poly2nk)
    return g22



# -*- coding: utf-8 -*-
"""
Database tables
"""

from banana.data.db import Base
from sqlalchemy import Boolean, Column, Integer, Text


class Operator(Base):  # pylint: disable=too-few-public-methods
    """
    Operator cards table
    """

    __tablename__ = "operators"

    interpolation_is_log = Column(Text)
    interpolation_polynomial_degree = Column(Integer)
    interpolation_xgrid = Column(Text)
    debug_skip_non_singlet = Column(Boolean)
    debug_skip_singlet = Column(Boolean)
    ev_op_max_order = Column(Integer)
    ev_op_iterations = Column(Integer)
    Q2grid = Column(Text)
    backward_inversion = Column(Text)



# -*- coding: utf-8 -*-
"""Zeta function values and other constants."""

import numpy as np
from scipy.special import zeta

zeta2 = zeta(2)
r""":math:`\zeta(2)`"""

zeta3 = zeta(3)
r""":math:`\zeta(3)`"""

zeta4 = zeta(4)
r""":math:`\zeta(4)`"""

zeta5 = zeta(5)
r""":math:`\zeta(5)`"""

log2 = np.log(2)
r""":math:`\ln(2)`"""

li4half = 0.517479
""":math:`Li_{4}(1/2)`"""



# -*- coding: utf-8 -*-
import copy
import math

from .genpdf import load


def create_info_file(theory_card, operators_card, num_members, info_update):
    """
    Generate a lhapdf info file from theory and operators card

    Parameters
    ----------
        theory_card : dict
            theory card
        operators_card : dict
            operators_card
        num_members : int
            number of pdf set members
        info_update : dict
            info to update

    Returns
    -------
        : dict
        info file in lhapdf format
    """
    template_info = copy.deepcopy(load.template_info)
    template_info["SetDesc"] = "Evolved PDF from " + str(theory_card["Q0"]) + " GeV"
    template_info["Authors"] = ""
    template_info["FlavorScheme"] = "variable"
    template_info.update(info_update)
    template_info["NumFlavors"] = 14
    template_info["Flavors"] = [-6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 21, 22]
    template_info["XMin"] = operators_card["interpolation_xgrid"][0]
    template_info["XMax"] = operators_card["interpolation_xgrid"][-1]
    template_info["NumMembers"] = num_members
    template_info["OrderQCD"] = theory_card["PTO"]
    template_info["QMin"] = round(math.sqrt(operators_card["Q2grid"][0]), 4)
    template_info["QMax"] = round(math.sqrt(operators_card["Q2grid"][-1]), 4)
    template_info["MZ"] = theory_card["MZ"]
    template_info["MUp"] = 0.0
    template_info["MDown"] = 0.0
    template_info["MStrange"] = 0.0
    template_info["MCharm"] = theory_card["mc"]
    template_info["MBottom"] = theory_card["mb"]
    template_info["MTop"] = theory_card["mt"]
    template_info["AlphaS_MZ"] = theory_card["alphas"]
    template_info["AlphaS_OrderQCD"] = theory_card["PTO"]
    # NB: Maybe I need the actual operator to obtain AlphaS_Qs
    return template_info



# -*- coding: utf-8 -*-
r"""
This module contains the |RGE| for the |MSbar| masses
"""
import numba as nb
import numpy as np
from scipy import integrate, optimize

from .basis_rotation import quark_names
from .beta import b_qcd, beta_qcd
from .couplings import Couplings, invert_matching_coeffs
from .gamma import gamma
from .thresholds import ThresholdsAtlas, flavor_shift, is_downward_path


def ker_exact(a0, a1, order, nf):
    r"""
    Exact |MSbar| |RGE| kernel

    Parameters
    ----------
        a0: float
            strong coupling at the initial scale
        a1: float
            strong coupling at the final scale
        oreder: tuple(int,int)
            perturbative order
        nf: int
            number of active flavours

    Returns
    -------
        ker_exact: float
            Exact |MSbar| kernel:

            .. math::
                k_{exact} = e^{-\int_{a_s(\mu_{h,0}^2)}^{a_s(\mu^2)}\gamma_m(a_s)/ \beta(a_s)da_s}
    """
    b_vec = [beta_qcd((2, 0), nf)]
    g_vec = [gamma(1, nf)]
    if order[0] >= 2:
        b_vec.append(beta_qcd((3, 0), nf))
        g_vec.append(gamma(2, nf))
    if order[0] >= 3:
        b_vec.append(beta_qcd((4, 0), nf))
        g_vec.append(gamma(3, nf))
    if order[0] >= 4:
        b_vec.append(beta_qcd((5, 0), nf))
        g_vec.append(gamma(4, nf))

    # quad ker
    def integrand(a, b_vec, g_vec):
        # minus sign goes away
        fgamma = np.sum([a**k * b for k, b in enumerate(g_vec)])
        fbeta = a * np.sum([a**k * b for k, b in enumerate(b_vec)])
        return fgamma / fbeta

    res = integrate.quad(
        integrand,
        a0,
        a1,
        args=(b_vec, g_vec),
        epsabs=1e-12,
        epsrel=1e-5,
        limit=100,
        full_output=1,
    )
    val, _ = res[:2]
    return np.exp(val)


@nb.njit(cache=True)
def ker_expanded(a0, a1, order, nf):
    r"""
    Expanded |MSbar| |RGE| kernel

    Parameters
    ----------
        a0: float
            strong coupling at the initial scale
        a1: float
            strong coupling at the final scale
        order: tuple(int,int)
            perturbative order
        nf: int
            number of active flavours

    Returns
    -------
        ker_expanded: float
            Expanded |MSbar| kernel:

            .. math::
                k_{expanded} &= \left (\frac{a_s(\mu^2)}{a_s(\mu_{h,0}^2)} \right )^{c_0}
                \frac{j_{exp}(a_s(\mu^2))}{j_{exp}(a_s(\mu_{h,0}^2))} \\
                j_{exp}(a_s) &= 1 + a_s \left [ c_1 - b_1 c_0 \right ] \\
                & + \frac{a_s^2}{2}
                \left [c_2 - c_1 b_1 - b_2 c_0 + b_1^2 c_0 + (c_1 - b_1 c_0)^2 \right ] \\
                & + \frac{a_s^3}{6} [ -2 b_3 c_0 - b_1^3 c_0 (1 + c_0) (2 + c_0) - 2 b_2 c_1 \\
                & - 3 b_2 c_0 c_1 + b_1^2 (2 + 3 c_0 (2 + c_0)) c_1 + c_1^3 + 3 c_1 c_2 \\
                & + b_1 (b_2 c_0 (4 + 3 c_0) - 3 (1 + c_0) c_1^2 - (2 + 3 c_0) c_2) + 2 c_3 ]
    """
    b0 = beta_qcd((2, 0), nf)
    c0 = gamma(1, nf) / b0
    ev_mass = np.power(a1 / a0, c0)
    num = 1.0
    den = 1.0
    if order[0] >= 2:
        b1 = b_qcd((3, 0), nf)
        c1 = gamma(2, nf) / b0
        u = c1 - b1 * c0
        num += a1 * u
        den += a0 * u
    if order[0] >= 3:
        b2 = b_qcd((4, 0), nf)
        c2 = gamma(3, nf) / b0
        u = (c2 - c1 * b1 - b2 * c0 + b1**2 * c0 + (c1 - b1 * c0) ** 2) / 2.0
        num += a1**2 * u
        den += a0**2 * u
    if order[0] >= 4:
        b3 = b_qcd((5, 0), nf)
        c3 = gamma(4, nf) / b0
        u = (
            1
            / 6
            * (
                -2 * b3 * c0
                - b1**3 * c0 * (1 + c0) * (2 + c0)
                - 2 * b2 * c1
                - 3 * b2 * c0 * c1
                + b1**2 * (2 + 3 * c0 * (2 + c0)) * c1
                + c1**3
                + 3 * c1 * c2
                + b1
                * (b2 * c0 * (4 + 3 * c0) - 3 * (1 + c0) * c1**2 - (2 + 3 * c0) * c2)
                + 2 * c3
            )
        )
        num += a1**3 * u
        den += a0**3 * u
    return ev_mass * num / den


def ker_dispatcher(q2_to, q2m_ref, strong_coupling, fact_to_ren, nf):
    r"""
    Select the |MSbar| kernel and compute the strong coupling values

    Parameters
    ----------
        q2_to: float
            final scale
        q2m_ref: float
            initial scale
        strong_coupling: eko.strong_coupling.StrongCoupling
            Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
            any q
        fact_to_ren: float
            factorization to renormalization scale ratio
        nf: int
            number of active flavours

    Returns
    -------
        ker:
            Expanded or exact |MSbar| kernel
    """
    a0 = strong_coupling.a(q2m_ref / fact_to_ren, q2m_ref, nf)[0]
    a1 = strong_coupling.a(q2_to / fact_to_ren, q2_to, nf)[0]
    method = strong_coupling.method
    order = strong_coupling.order
    if method == "expanded":
        return ker_expanded(a0, float(a1), order, nf)
    return ker_exact(a0, a1, order, nf)


def compute_matching_coeffs_up(nf):
    """
    |MSbar| matching coefficients :cite:`Liu:2015fxa` at threshold
    when moving to a regime with *more* flavors.

    Note :cite:`Liu:2015fxa` (eq 5.) reports the backward relation,
    multiplied by a factor of 4 (and 4^2 ...)

    Parameters
    ----------
        nf:
            number of active flavors in the lower patch

    Returns
    -------
        matching_coeffs_up:
            downward matching coefficient matrix
    """
    matching_coeffs_up = np.zeros((4, 4))
    matching_coeffs_up[2, 0] = -89.0 / 27.0
    matching_coeffs_up[2, 1] = 20.0 / 9.0
    matching_coeffs_up[2, 2] = -4.0 / 3.0

    matching_coeffs_up[3, 0] = -118.248 - 1.58257 * nf
    matching_coeffs_up[3, 1] = 71.7887 + 7.85185 * nf
    matching_coeffs_up[3, 2] = -700.0 / 27.0
    matching_coeffs_up[3, 3] = -232.0 / 27.0 + 16.0 / 27.0 * nf

    return matching_coeffs_up


def compute_matching_coeffs_down(nf):
    """
    |MSbar| matching coefficients :cite:`Liu:2015fxa` (eq 5) at threshold
    when moving to a regime with *less* flavors.

    Parameters
    ----------
        nf:
            number of active flavors in the lower patch

    Returns
    -------
        matching_coeffs_down:
            downward matching coefficient matrix
    """
    c_up = compute_matching_coeffs_up(nf)
    return invert_matching_coeffs(c_up)


def solve(
    m2_ref,
    q2m_ref,
    strong_coupling,
    nf_ref,
    fact_to_ren,
):
    r"""
    Compute the |MSbar| mass solving the equation :math:`m_{\overline{MS}}(m) = m`
    for a fixed number of nf

    Parameters
    ----------
        m2_ref: float
            squared initial mass reference
        q2m_ref: float
            squared initial scale
        strong_coupling: eko.strong_coupling.StrongCoupling
            Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
            any q
        nf_ref: int
            number of active flavours at the scale q2m_ref, where the solution is searched
        fact_to_ren: float
            :math:`\mu_F^2/\mu_R^2`

    Returns
    -------
        m2 : float
            :math:`m_{\overline{MS}}(\mu_2)^2`
    """

    def rge(m2, q2m_ref, strong_coupling, fact_to_ren, nf_ref):
        return (
            m2_ref
            * ker_dispatcher(m2, q2m_ref, strong_coupling, fact_to_ren, nf_ref) ** 2
            - m2
        )

    msbar_mass = optimize.fsolve(
        rge, q2m_ref, args=(q2m_ref, strong_coupling, fact_to_ren, nf_ref)
    )
    return float(msbar_mass)


def evolve(
    m2_ref, q2m_ref, strong_coupling, fact_to_ren, q2_to, nf_ref=None, nf_to=None
):
    r"""
    Perform the |MSbar| mass evolution up to given scale.
    It allows for different number of active flavors.

    Parameters
    ----------
        m2_ref: float
            squared initial mass reference
        q2m_ref: float
            squared initial scale
        strong_coupling: eko.strong_coupling.StrongCoupling
            Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
            any q
        fact_to_ren: float
            :math:`\mu_F^2/\mu_R^2`
        q2_to: float
            scale at which the mass is computed
        nf_ref: int
            number of flavor active at the reference scale
        nf_to: int
            number of flavor active at the target scale

    Returns
    -------
        m2 : float
            :math:`m_{\overline{MS}}(\mu_2)^2`
    """
    thr_atlas = ThresholdsAtlas(
        np.array(strong_coupling.thresholds.area_walls)[1:-1],
        q2m_ref,
        nf_ref,
        strong_coupling.thresholds.thresholds_ratios,
    )
    path = thr_atlas.path(q2_to, nf_to)
    is_downward = is_downward_path(path)
    shift = flavor_shift(is_downward)

    ev_mass = 1.0
    for k, seg in enumerate(path):
        # skip a very short segment, but keep the matching
        ker_evol = 1.0
        if not np.isclose(seg.q2_from, seg.q2_to):
            ker_evol = (
                ker_dispatcher(
                    seg.q2_to, seg.q2_from, strong_coupling, fact_to_ren, seg.nf
                )
                ** 2
            )
        # apply matching condition
        if k < len(path) - 1:
            # TODO: do we need to add np.log(fac_to_ren) here ???
            L = np.log(thr_atlas.thresholds_ratios[seg.nf - shift])
            m_coeffs = (
                compute_matching_coeffs_down(seg.nf - 1)
                if is_downward
                else compute_matching_coeffs_up(seg.nf)
            )
            matching = 1.0
            for pto in range(1, strong_coupling.order[0]):
                for l in range(pto + 1):
                    as_thr = strong_coupling.a(
                        seg.q2_to / fact_to_ren, seg.q2_to, seg.nf - shift + 4
                    )[0]
                    matching += as_thr**pto * L**l * m_coeffs[pto, l]
            ker_evol *= matching
        ev_mass *= ker_evol
    return m2_ref * ev_mass


def compute(theory_card):
    r"""
    Compute the |MSbar| masses solving the equation :math:`m_{\bar{MS}}(\mu) = \mu`
    for each heavy quark and consistent boundary contitions.

    Parameters
    ----------
        theory_card: dict
            theory run card

    Returns
    -------
        masses: list
            list of |MSbar| masses squared
    """
    # TODO: sketch in the docs how the MSbar computation works with a figure.
    nfa_ref = theory_card["nfref"]

    q2_ref = np.power(theory_card["Qref"], 2)
    masses = np.concatenate((np.zeros(nfa_ref - 3), np.full(6 - nfa_ref, np.inf)))
    fact_to_ren = theory_card["fact_to_ren_scale_ratio"] ** 2

    def sc(thr_masses):
        return Couplings.from_dict(theory_card, masses=thr_masses)

    # First you need to look for the thr around the given as_ref
    heavy_quarks = quark_names[3:]
    hq_idxs = np.arange(0, 3)
    if nfa_ref > 4:
        heavy_quarks = reversed(heavy_quarks)
        hq_idxs = reversed(hq_idxs)

    # loop on heavy quarks and compute the msbar masses
    for q_idx, hq in zip(hq_idxs, heavy_quarks):
        q2m_ref = np.power(theory_card[f"Qm{hq}"], 2)
        m2_ref = np.power(theory_card[f"m{hq}"], 2)

        # check if mass is already given at the pole -> done
        if q2m_ref == m2_ref:
            masses[q_idx] = m2_ref
            continue

        # update the alphas thr scales
        nf_target = q_idx + 3
        shift = -1

        # check that alphas is given with a consistent number of flavors
        if q_idx + 4 == nfa_ref and q2m_ref > q2_ref:
            raise ValueError(
                f"In MSBAR scheme, Qm{hq} should be lower than Qref, \
                if alpha_s is given with nfref={nfa_ref} at scale Qref={q2_ref}"
            )
        if q_idx + 4 == nfa_ref + 1 and q2m_ref < q2_ref:
            raise ValueError(
                f"In MSBAR scheme, Qm{hq} should be greater than Qref, \
                if alpha_s is given with nfref={nfa_ref} at scale Qref={q2_ref}"
            )

        # check that for higher patches you do forward running
        # with consistent conditions
        if q_idx + 3 >= nfa_ref and q2m_ref >= m2_ref:
            raise ValueError(
                f"In MSBAR scheme, Qm{hq} should be lower than m{hq} \
                        if alpha_s is given with nfref={nfa_ref} at scale Qref={q2_ref}"
            )

        # check that for lower patches you do backward running
        # with consistent conditions
        if q_idx + 3 < nfa_ref:
            if q2m_ref < m2_ref:
                raise ValueError(
                    f"In MSBAR scheme, Qm{hq} should be greater than m{hq} \
                        if alpha_s is given with nfref={nfa_ref} at scale Qref={q2_ref}"
                )
            nf_target += 1
            shift = 1

        # if the initial condition is not in the target patch,
        # you need to evolve it until nf_target patch wall is reached:
        #   for backward you reach the higher, for forward the lower.
        # len(masses[q2m_ref > masses]) + 3 is the nf at the given reference scale
        nf_ref = len(masses[q2m_ref > masses]) + 3
        if nf_target != nf_ref:
            q2_to = masses[q_idx + shift]
            m2_ref = evolve(
                m2_ref,
                q2m_ref,
                sc(masses),
                fact_to_ren,
                q2_to,
                nf_ref=nf_ref,
                nf_to=nf_target,
            )
            q2m_ref = q2_to

        # now solve the RGE
        masses[q_idx] = solve(
            m2_ref,
            q2m_ref,
            sc(masses),
            nf_target,
            fact_to_ren,
        )

    # Check the msbar ordering
    if not (masses == np.sort(masses)).all():
        raise ValueError("Msbar masses are not to be sorted")
    return masses



# -*- coding: utf-8 -*-
"""
This file contains the output management
"""
import io
import logging
import pathlib
import tarfile
import tempfile
import warnings

import lz4.frame
import numpy as np
import yaml

from . import basis_rotation as br
from . import interpolation, version

logger = logging.getLogger(__name__)


class Output(dict):
    """
    Wrapper for the output to help with application
    to PDFs and dumping to file.
    """

    def xgrid_reshape(self, targetgrid=None, inputgrid=None):
        """
        Changes the operators to have in the output targetgrid and/or in the input inputgrid.

        The operation is in-place.

        Parameters
        ----------
            targetgrid : None or list
                xgrid for the target
            inputgrid : None or list
                xgrid for the input
        """
        # calling with no arguments is an error
        if targetgrid is None and inputgrid is None:
            raise ValueError("Nor inputgrid nor targetgrid was given")
        # now check to the current status
        if (
            targetgrid is not None
            and len(targetgrid) == len(self["targetgrid"])
            and np.allclose(targetgrid, self["targetgrid"])
        ):
            targetgrid = None
            warnings.warn("The new targetgrid is close to the current targetgrid")
        if (
            inputgrid is not None
            and len(inputgrid) == len(self["inputgrid"])
            and np.allclose(inputgrid, self["inputgrid"])
        ):
            inputgrid = None
            warnings.warn("The new inputgrid is close to the current inputgrid")
        # after the checks: if there is still nothing to do, skip
        if targetgrid is None and inputgrid is None:
            logger.debug("Nothing done.")
            return

        # construct matrices
        if targetgrid is not None:
            b = interpolation.InterpolatorDispatcher(
                self["targetgrid"],
                self["interpolation_polynomial_degree"],
                self["interpolation_is_log"],
                False,
            )
            target_rot = b.get_interpolation(targetgrid)
            self["targetgrid"] = np.array(targetgrid)
        if inputgrid is not None:
            b = interpolation.InterpolatorDispatcher(
                inputgrid,
                self["interpolation_polynomial_degree"],
                self["interpolation_is_log"],
                False,
            )
            input_rot = b.get_interpolation(self["inputgrid"])
            self["inputgrid"] = np.array(inputgrid)

        # build new grid
        for elem in self["Q2grid"].values():
            ops = elem["operators"]
            errs = elem["operator_errors"]
            if targetgrid is not None and inputgrid is None:
                ops = np.einsum("ij,ajbk->aibk", target_rot, ops)
                errs = np.einsum("ij,ajbk->aibk", target_rot, errs)
            elif inputgrid is not None and targetgrid is None:
                ops = np.einsum("ajbk,kl->ajbl", ops, input_rot)
                errs = np.einsum("ajbk,kl->ajbl", errs, input_rot)
            else:
                ops = np.einsum("ij,ajbk,kl->aibl", target_rot, ops, input_rot)
                errs = np.einsum("ij,ajbk,kl->aibl", target_rot, errs, input_rot)
            elem["operators"] = ops
            elem["operator_errors"] = errs

    def flavor_reshape(self, targetbasis=None, inputbasis=None):
        """
        Changes the operators to have in the output targetbasis and/or in the input inputbasis.

        The operation is in-place.

        Parameters
        ----------
            targetbasis : numpy.ndarray
                target rotation specified in the flavor basis
            inputbasis : None or list
                input rotation specified in the flavor basis
        """
        # calling with no arguments is an error
        if targetbasis is None and inputbasis is None:
            raise ValueError("Nor inputbasis nor targetbasis was given")
        # now check to the current status
        if targetbasis is not None and np.allclose(
            targetbasis, np.eye(len(self["targetpids"]))
        ):
            targetbasis = None
            warnings.warn("The new targetbasis is close to current basis")
        if inputbasis is not None and np.allclose(
            inputbasis, np.eye(len(self["inputpids"]))
        ):
            inputbasis = None
            warnings.warn("The new inputbasis is close to current basis")
        # after the checks: if there is still nothing to do, skip
        if targetbasis is None and inputbasis is None:
            logger.debug("Nothing done.")
            return

        # flip input around
        if inputbasis is not None:
            inv_inputbasis = np.linalg.inv(inputbasis)

        # build new grid
        for elem in self["Q2grid"].values():
            ops = elem["operators"]
            errs = elem["operator_errors"]
            if targetbasis is not None and inputbasis is None:
                ops = np.einsum("ca,ajbk->cjbk", targetbasis, ops)
                errs = np.einsum("ca,ajbk->cjbk", targetbasis, errs)
            elif inputbasis is not None and targetbasis is None:
                ops = np.einsum("ajbk,bd->ajdk", ops, inv_inputbasis)
                errs = np.einsum("ajbk,bd->ajdk", errs, inv_inputbasis)
            else:
                ops = np.einsum("ca,ajbk,bd->cjdk", targetbasis, ops, inv_inputbasis)
                errs = np.einsum("ca,ajbk,bd->cjdk", targetbasis, errs, inv_inputbasis)
            elem["operators"] = ops
            elem["operator_errors"] = errs
        # drop PIDs - keeping them int nevertheless
        if inputbasis is not None:
            self["inputpids"] = [0] * len(self["inputpids"])
        if targetbasis is not None:
            self["targetpids"] = [0] * len(self["targetpids"])

    def to_evol(self, source=True, target=False):
        """
        Rotate the operator into evolution basis.

        This also assigns also the pids. The operation is in-place.

        Parameters
        ----------
            source : bool
                rotate on the input tensor
            target : bool
                rotate on the output tensor
        """
        # rotate
        inputbasis = br.rotate_flavor_to_evolution if source else None
        targetbasis = br.rotate_flavor_to_evolution if target else None
        self.flavor_reshape(inputbasis=inputbasis, targetbasis=targetbasis)
        # assign pids
        if source:
            self["inputpids"] = br.evol_basis_pids
        if target:
            self["targetpids"] = br.evol_basis_pids

    def get_raw(self, binarize=True, skip_q2_grid=False):
        """
        Serialize result as dict/YAML.

        This maps the original numpy matrices to lists.

        Parameters
        ----------
            binarize : bool
                dump in binary format (instead of list format)

        Returns
        -------
            out : dict
                dictionary which will be written on output
        """
        # prepare output dict
        out = {"Q2grid": {}, "eko_version": version.__version__}
        # dump raw elements
        for f in [
            "interpolation_polynomial_degree",
            "interpolation_is_log",
            "q2_ref",
        ]:
            out[f] = self[f]

        # list() work both for tuple and list
        out["inputpids"] = list(self["inputpids"])
        out["targetpids"] = list(self["targetpids"])
        # make raw lists
        # TODO: is interpolation_xgrid really needed in the output?
        for k in ["interpolation_xgrid", "targetgrid", "inputgrid"]:
            out[k] = self[k].tolist()
        # make operators raw
        if not skip_q2_grid:
            for q2, op in self["Q2grid"].items():
                out["Q2grid"][q2] = {}
                for k, v in op.items():
                    if k == "alphas":
                        out["Q2grid"][q2][k] = float(v)
                        continue
                    if binarize:
                        out["Q2grid"][q2][k] = lz4.frame.compress(v.tobytes())
                    else:
                        out["Q2grid"][q2][k] = v.tolist()
        else:
            out["Q2grid"] = self["Q2grid"]
        return out

    def dump_yaml(self, stream=None, binarize=True, skip_q2_grid=False):
        """
        Serialize result as YAML.

        Parameters
        ----------
            stream : None or stream
                if given, dump is written on it
            binarize : bool
                dump in binary format (instead of list format)
            skip_q2_grid : bool
                avoid dumping Q2grid (i.e. the actual operators) into the yaml
                file (default: ``False``)

        Returns
        -------
            dump : any
                result of dump(output, stream), i.e. a string, if no stream is given or
                Null, if written successfully to stream
        """
        # TODO explicitly silence yaml
        out = self.get_raw(binarize, skip_q2_grid=skip_q2_grid)
        return yaml.dump(out, stream)

    def dump_yaml_to_file(self, filename, binarize=True, skip_q2_grid=False):
        """
        Writes YAML representation to a file.

        Parameters
        ----------
            filename : str
                target file name
            binarize : bool
                dump in binary format (instead of list format)
            skip_q2_grid : bool
                avoid dumping Q2grid (i.e. the actual operators) into the yaml
                file (default: ``False``)

        Returns
        -------
            ret : any
                result of dump(output, stream), i.e. Null if written successfully
        """
        with open(filename, "w", encoding="utf-8") as f:
            ret = self.dump_yaml(f, binarize, skip_q2_grid=skip_q2_grid)
        return ret

    def dump_tar(self, tarname):
        """
        Writes representation into a tar archive containing:

        - metadata (in YAML)
        - operator (in numpy ``.npy`` format)

        Parameters
        ----------
            tarname : str
                target file name
        """
        tarpath = pathlib.Path(tarname)
        if tarpath.suffix != ".tar":
            raise ValueError(f"'{tarname}' is not a valid tar filename, wrong suffix")

        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            cls = self.__class__
            metadata = cls(**{str(k): v for k, v in self.items() if k != "Q2grid"})
            metadata["Q2grid"] = list(self["Q2grid"].keys())

            yamlname = tmpdir / "metadata.yaml"
            metadata.dump_yaml_to_file(yamlname, skip_q2_grid=True)

            for kind in next(iter(self["Q2grid"].values())).keys():
                operator = np.stack([q2[kind] for q2 in self["Q2grid"].values()])
                stream = io.BytesIO()
                np.save(stream, operator)
                stream.seek(0)
                with lz4.frame.open(
                    (tmpdir / kind).with_suffix(".npy.lz4"), "wb"
                ) as fo:
                    fo.write(stream.read())

            with tarfile.open(tarpath, "w") as tar:
                tar.add(tmpdir, arcname=tarpath.stem)

    @classmethod
    def load_yaml(cls, stream, skip_q2_grid=False):
        """
        Load YAML representation from stream

        Parameters
        ----------
            stream : any
                source stream
            skip_q2_grid : bool
                avoid loading Q2grid (i.e. the actual operators) from the yaml
                file (default: ``False``)

        Returns
        -------
            obj : output
                loaded object
        """
        obj = yaml.safe_load(stream)
        len_tpids = len(obj["targetpids"])
        len_ipids = len(obj["inputpids"])
        len_tgrid = len(obj["targetgrid"])
        len_igrid = len(obj["inputgrid"])
        # cast lists to numpy
        for k in ["interpolation_xgrid", "inputgrid", "targetgrid"]:
            obj[k] = np.array(obj[k])
        # make operators numpy
        if not skip_q2_grid:
            for op in obj["Q2grid"].values():
                for k, v in op.items():
                    if k == "alphas":
                        v = float(v)
                    elif isinstance(v, list):
                        v = np.array(v)
                    elif isinstance(v, bytes):
                        v = np.frombuffer(lz4.frame.decompress(v))
                        v = v.reshape(len_tpids, len_tgrid, len_ipids, len_igrid)
                    op[k] = v
        return cls(obj)

    @classmethod
    def load_yaml_from_file(cls, filename, skip_q2_grid=False):
        """
        Load YAML representation from file

        Parameters
        ----------
            filename : str
                source file name
            skip_q2_grid : bool
                avoid loading Q2grid (i.e. the actual operators) from the yaml
                file (default: ``False``)

        Returns
        -------
            obj : output
                loaded object
        """
        obj = None
        with open(filename, encoding="utf-8") as o:
            obj = Output.load_yaml(o, skip_q2_grid)
        return obj

    @classmethod
    def load_tar(cls, tarname):
        """
        Load tar representation from file (compliant with :meth:`dump_tar`
        output).

        Parameters
        ----------
            tarname : str
                source tar name

        Returns
        -------
            obj : output
                loaded object
        """
        tarpath = pathlib.Path(tarname)

        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            with tarfile.open(tarpath, "r") as tar:
                tar.extractall(tmpdir)

            # metadata = cls(**{str(k): v for k, v in self.items() if k != "Q2grid"})
            # metadata["Q2grid"] = list(self["Q2grid"].keys())

            innerdir = list(tmpdir.glob("*"))[0]
            yamlname = innerdir / "metadata.yaml"
            metadata = cls.load_yaml_from_file(yamlname, skip_q2_grid=True)

            grids = {}
            for fp in innerdir.glob("*.npy.lz4"):
                with lz4.frame.open(fp, "rb") as fd:
                    stream = io.BytesIO(fd.read())
                    stream.seek(0)
                    grids[pathlib.Path(fp.stem).stem] = np.load(stream)

                fp.unlink()

            q2grid = metadata["Q2grid"]
            operator_grid = {}
            for q2, slices in zip(q2grid, zip(*grids.values())):
                operator_grid[q2] = dict(zip(grids.keys(), slices))
            metadata["Q2grid"] = operator_grid

        return metadata



# -*- coding: utf-8 -*-
import numpy as np

from eko import basis_rotation as br
from eko import interpolation


def apply_pdf(output, lhapdf_like, targetgrid=None, rotate_to_evolution_basis=False):
    """
    Apply all available operators to the input PDFs.

    Parameters
    ----------
        output : eko.output.Output
            eko output object containing all informations
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
            and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
        targetgrid : list
            if given, interpolates to the pdfs given at targetgrid (instead of xgrid)
        rotate_to_evolution_basis : bool
            if True rotate to evoluton basis

    Returns
    -------
        out_grid : dict
            output PDFs and their associated errors for the computed Q2grid
    """
    if rotate_to_evolution_basis:
        return apply_pdf_flavor(
            output, lhapdf_like, targetgrid, br.rotate_flavor_to_evolution
        )
    return apply_pdf_flavor(output, lhapdf_like, targetgrid)


def apply_pdf_flavor(output, lhapdf_like, targetgrid=None, flavor_rotation=None):
    """
    Apply all available operators to the input PDFs.

    Parameters
    ----------
        output : eko.output.Output
            eko output object containing all informations
        lhapdf_like : object
            object that provides an xfxQ2 callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
            and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
        targetgrid : list
            if given, interpolates to the pdfs given at targetgrid (instead of xgrid)
        flavor_rotation : np.ndarray
            Rotation matrix in flavor space

    Returns
    -------
        out_grid : dict
            output PDFs and their associated errors for the computed Q2grid
    """
    # create pdfs
    pdfs = np.zeros((len(output["inputpids"]), len(output["inputgrid"])))
    for j, pid in enumerate(output["inputpids"]):
        if not lhapdf_like.hasFlavor(pid):
            continue
        pdfs[j] = np.array(
            [
                lhapdf_like.xfxQ2(pid, x, output["q2_ref"]) / x
                for x in output["inputgrid"]
            ]
        )

    # build output
    out_grid = {}
    for q2, elem in output["Q2grid"].items():
        pdf_final = np.einsum("ajbk,bk", elem["operators"], pdfs)
        error_final = np.einsum("ajbk,bk", elem["operator_errors"], pdfs)
        out_grid[q2] = {
            "pdfs": dict(zip(output["targetpids"], pdf_final)),
            "errors": dict(zip(output["targetpids"], error_final)),
        }

    # rotate to evolution basis
    if flavor_rotation is not None:
        for q2, op in out_grid.items():
            pdf = flavor_rotation @ np.array(
                [op["pdfs"][pid] for pid in br.flavor_basis_pids]
            )
            errors = flavor_rotation @ np.array(
                [op["errors"][pid] for pid in br.flavor_basis_pids]
            )
            op["pdfs"] = dict(zip(br.evol_basis, pdf))
            op["errors"] = dict(zip(br.evol_basis, errors))

    # rotate/interpolate to target grid
    if targetgrid is not None:
        b = interpolation.InterpolatorDispatcher.from_dict(output, False)
        rot = b.get_interpolation(targetgrid)
        for evpdf in out_grid.values():
            for pdf_label in evpdf["pdfs"]:
                evpdf["pdfs"][pdf_label] = np.matmul(rot, evpdf["pdfs"][pdf_label])
                evpdf["errors"][pdf_label] = np.matmul(rot, evpdf["errors"][pdf_label])

    return out_grid



# -*- coding: utf-8 -*-
import io
import pathlib
import re

import numpy as np
import yaml


def list_to_str(ls, fmt="%.6e"):
    """
    Convert a list of numbers to a string

    Parameters
    ----------
        ls : list(float)
            list
        fmt : str
            format string

    Returns
    -------
        str :
            final string
    """
    return " ".join([fmt % x for x in ls])


def array_to_str(ar):
    """
    Convert an array of numbers to a string

    Parameters
    ----------
        ar : list(list(float))
            array

    Returns
    -------
        str :
            final string
    """
    table = ""
    for line in ar:
        table += f"{line[0]:.8e} " + list_to_str(line[1:], fmt="%.8e") + "\n"
    return table


def dump_blocks(name, member, blocks, pdf_type=None):
    """
    Write LHAPDF data file.

    Parameters
    ----------
        name : str or os.PathLike
            target name or path
        member : int
            PDF member
        blocks : list(dict)
            pdf blocks of data
        inherit : str
            str to be copied in the head of member files
    """
    path_name = pathlib.Path(name)
    target = path_name / f"{path_name.stem}_{member:04d}.dat"
    target.parent.mkdir(exist_ok=True)
    with open(target, "w", encoding="utf-8") as o:
        if pdf_type is None:
            if member == 0:
                o.write("PdfType: central\n")
            else:
                o.write("PdfType: replica\n")
        else:
            o.write(pdf_type)
        o.write("Format: lhagrid1\n---\n")
        for b in blocks:
            o.write(list_to_str(b["xgrid"]) + "\n")
            o.write(list_to_str(list(np.sqrt(b["Q2grid"]))) + "\n")
            o.write(list_to_str(b["pids"], "%d") + "\n")
            o.write(array_to_str(b["data"]))
            o.write("---\n")


def dump_info(name, info):
    """
    Write LHAPDF info file.

    NOTE: Since LHAPDF info files are not truely yaml files,
    we have to use a slightly more complicated function to
    dump the info file.

    Parameters
    ----------
        name : str or os.Pathlike
            target name or path
        info : dict
            info dictionary
    """
    path_name = pathlib.Path(name)
    target = path_name / f"{path_name.stem}.info"
    target.parent.mkdir(exist_ok=True)
    # write on string stream to capture output
    stream = io.StringIO()
    yaml.safe_dump(info, stream, default_flow_style=True, width=100000, line_break="\n")
    cnt = stream.getvalue()
    # now insert some newlines for each key
    new_cnt = re.sub(r", ([A-Za-z_]+\d?):", r"\n\1:", cnt.strip()[1:-1])
    with open(target, "w", encoding="utf-8") as o:
        o.write(new_cnt)


def dump_set(name, info, member_blocks, pdf_type_list=None):
    """
    Dump a whole set.

    Parameters
    ----------
        name : str
            target name
        info : dict
            info dictionary
        member_blocks : list(list(dict))
            blocks for all members
        pdf_type : list(str)
            list of strings to be copied in the head of member files
    """
    dump_info(name, info)
    for mem, blocks in enumerate(member_blocks):
        if not isinstance(pdf_type_list, list):
            dump_blocks(name, mem, blocks)
        elif len(pdf_type_list) == 0:
            dump_blocks(name, mem, blocks)
        else:
            dump_blocks(name, mem, blocks, pdf_type=pdf_type_list[mem])



# -*- coding: utf-8 -*-
r"""
Integrals needed for the exact evolutions are given by:

.. math::
    j^{(n,m)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s}\!da_s'\,\frac{(a_s')^{1+n}}{-\beta^{(m)}(a_s')}

The expanded integrals are obtained from the exact results by Taylor expanding in the limit
:math:`a_s,a_s^{0} \to 0` until :math:`\mathcal{O}( a_s^{m+1})` for :math:`N^{m}LO` computations.
"""

import numba as nb
import numpy as np

from .. import beta


@nb.njit(cache=True)
def j00(a1, a0, nf):
    r"""
    LO-LO exact evolution integral.

    .. math::
        j^{(0,0)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s} \frac{da_s'}{\beta_0 a_s'}
           = \frac{\ln(a_s/a_s^0)}{\beta_0}

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j00 : float
            integral
    """
    return np.log(a1 / a0) / beta.beta_qcd((2, 0), nf)


@nb.njit(cache=True)
def j11_exact(a1, a0, nf):
    r"""
    NLO-NLO exact evolution integral.

    .. math::
        j^{(1,1)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s}\!da_s'\,
                                \frac{a_s'^2}{\beta_0 a_s'^2 + \beta_1 a_s'^3}
            = \frac{1}{\beta_1}\ln\left(\frac{1+b_1 a_s}{1+b_1 a_s^0}\right)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j11 : float
            integral
    """
    beta_qcd_as3 = beta.beta_qcd((3, 0), nf)
    b1 = beta.b_qcd((3, 0), nf)
    return (1.0 / beta_qcd_as3) * np.log((1.0 + a1 * b1) / (1.0 + a0 * b1))


@nb.njit(cache=True)
def j11_expanded(a1, a0, nf):
    r"""
    NLO-NLO expanded evolution integral.

    .. math::
        j^{(1,1)}_{exp}(a_s,a_s^0) = \frac 1 {\beta_0}(a_s - a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j11_exp : float
            integral
    """
    return 1.0 / beta.beta_qcd((2, 0), nf) * (a1 - a0)


@nb.njit(cache=True)
def j01_exact(a1, a0, nf):
    r"""
    LO-NLO exact evolution integral.

    .. math::
        j^{(0,1)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s}\!da_s'\,
                            \frac{a_s'}{\beta_0 a_s'^2 + \beta_1 a_s'^3}
               = j^{(0,0)}(a_s,a_s^0) - b_1 j^{(1,1)}(a_s,a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j11 : float
            integral
    """
    return j00(a1, a0, nf) - beta.b_qcd((3, 0), nf) * j11_exact(a1, a0, nf)


@nb.njit(cache=True)
def j01_expanded(a1, a0, nf):
    r"""
    LO-NLO expanded evolution integral.

    .. math::
        j^{(0,1)}_{exp}(a_s,a_s^0) = j^{(0,0)}(a_s,a_s^0) - b_1 j^{(1,1)}_{exp}(a_s,a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j01_exp : float
            integral
    """
    return j00(a1, a0, nf) - beta.b_qcd((3, 0), nf) * j11_expanded(a1, a0, nf)


@nb.njit(cache=True)
def j22_exact(a1, a0, nf):
    r"""
    NNLO-NNLO exact evolution integral.

    .. math::
        j^{(2,2)}(a_s,a_s^0) &=
            \int\limits_{a_s^0}^{a_s}\!da_s'\,\frac{a_s'^3}
                        {\beta_0 a_s'^2 + \beta_1 a_s'^3 + \beta_2 a_s'^4}
            = \frac{1}{\beta_2}\ln\left(
                    \frac{1 + a_s ( b_1 + b_2 a_s ) }{ 1 + a_s^0 ( b_1 + b_2 a_s^0 )}\right)
                - \frac{b_1 \delta}{ \beta_2 \Delta} \\
            \delta &= \atan \left( \frac{b_1 + 2 a_s b_2 }{ \Delta} \right)
                    - \atan \left( \frac{b_1 + 2 a_s^0 b_2 }{ \Delta} \right) \\
            \Delta &= \sqrt{4 b_2 - b_1^2}

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j22 : complex
            integral
    """
    b1 = beta.b_qcd((3, 0), nf)
    b2 = beta.b_qcd((4, 0), nf)
    # allow Delta to be complex for nf = 6, the final result will be real
    Delta = np.sqrt(complex(4 * b2 - b1**2))
    delta = np.arctan((b1 + 2 * a1 * b2) / Delta) - np.arctan(
        (b1 + 2 * a0 * b2) / Delta
    )
    log = np.log((1 + a1 * (b1 + b2 * a1)) / (1 + a0 * (b1 + b2 * a0)))
    return 1 / (2 * beta.beta_qcd((4, 0), nf)) * log - b1 / (
        beta.beta_qcd((4, 0), nf)
    ) * np.real(delta / Delta)


@nb.njit(cache=True)
def j12_exact(a1, a0, nf):
    r"""
    NLO-NNLO exact evolution integral.

    .. math::
        j^{(1,2)}(a_s,a_s^0) &= \int\limits_{a_s^0}^{a_s}\!da_s'\,\frac{a_s'^2}{\beta_0 a_s'^2 + \beta_1 a_s'^3 + \beta_2 a_s'^4}\\
               &= \frac{2 \delta}{\beta_0 \Delta}  \\
        \delta &= \atan \left( \frac{b_1 + 2 a_s b_2 }{ \Delta} \right) - \atan \left( \frac{b_1 + 2 a_s^0 b_2 }{ \Delta} \right) \\
        \Delta &= \sqrt{4 b_2 - b_1^2}

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j12 : complex
            integral
    """  # pylint: disable=line-too-long
    b1 = beta.b_qcd((3, 0), nf)
    b2 = beta.b_qcd((4, 0), nf)
    # allow Delta to be complex for nf = 6, the final result will be real
    Delta = np.sqrt(complex(4 * b2 - b1**2))
    delta = np.arctan((b1 + 2 * a1 * b2) / Delta) - np.arctan(
        (b1 + 2 * a0 * b2) / Delta
    )
    return 2.0 / (beta.beta_qcd((2, 0), nf)) * np.real(delta / Delta)


@nb.njit(cache=True)
def j02_exact(a1, a0, nf):
    r"""
    LO-NNLO exact evolution integral.

    .. math::
        j^{(0,2)}(a_s,a_s^0) &= \int\limits_{a_s^0}^{a_s}\!da_s'\,
              \frac{a_s'}{\beta_0 a_s'^2 + \beta_1 a_s'^3 + \beta_2 a_s'^4}\\
            &= j^{(0,0)}(a_s,a_s^0) - b_1 j^{(1,2)}(a_s,a_s^0) - b_2 j^{(2,2)}(a_s,a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j02 : complex
            integral
    """
    return (
        j00(a1, a0, nf)
        - beta.b_qcd((3, 0), nf) * j12_exact(a1, a0, nf)
        - beta.b_qcd((4, 0), nf) * j22_exact(a1, a0, nf)
    )


@nb.njit(cache=True)
def j22_expanded(a1, a0, nf):
    r"""
    NNLO-NNLO expanded evolution integral.

    .. math::
        j^{(2,2)}_{exp}(a_s,a_s^0) = \frac{1}{2 \beta_0} \left( a_s^2 -  (a_s^0)^{2} \right)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j22_exp : float
            integral
    """
    return 1 / (2 * beta.beta_qcd((2, 0), nf)) * (a1**2 - a0**2)


@nb.njit(cache=True)
def j12_expanded(a1, a0, nf):
    r"""
    NLO-NNLO expanded evolution integral.

    .. math::
        j^{(1,2)}_{exp}(a_s,a_s^0) = \frac{1}{\beta_0}\left[ a_s - a_s^0 -
                      \frac{b_1}{2} \left( a_s^2 - (a_s^0)^{2} \right)\right]

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j12_exp : float
            integral
    """
    b1 = beta.b_qcd((3, 0), nf)
    return 1 / beta.beta_qcd((2, 0), nf) * (a1 - a0 - b1 / 2 * (a1**2 - a0**2))


@nb.njit(cache=True)
def j02_expanded(a1, a0, nf):
    r"""
    LO-NNLO expanded evolution integral.

    .. math::
        j^{(0,2)}_{exp}(a_s,a_s^0) = j^{(0,0)}(a_s,a_s^0) - b_1 j^{(1,2)}_{exp}(a_s,a_s^0)
                                      - b_2 j^{(2,2)}_{exp}(a_s,a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j02_exp : float
            integral
    """
    return (
        j00(a1, a0, nf)
        - beta.b_qcd((3, 0), nf) * j12_expanded(a1, a0, nf)
        - beta.b_qcd((4, 0), nf) * j22_expanded(a1, a0, nf)
    )



# -*- coding: utf-8 -*-
r"""Implementation of Mellin transformation of logarithms.

We provide transforms of:

- :math:`(1-x)\ln^k(1-x), \quad k = 1,2,3`
- :math:`\ln^k(1-x), \quad k = 1,3,4,5`

"""
import numba as nb

from .constants import zeta3


@nb.njit(cache=True)
def lm11m1(n, S1):
    r"""Mellin transform of :math:`(1-x)\ln(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln(1-x)](N)`
    """
    return 1 / (1 + n) ** 2 - S1 / (1 + n) ** 2 - S1 / (n * (1 + n) ** 2)


@nb.njit(cache=True)
def lm12m1(n, S1, S2):
    r"""Mellin transform of :math:`(1-x)\ln^2(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln^2(1-x)](N)`

    """
    return (
        -2 / (1 + n) ** 3
        - (2 * S1) / (1 + n) ** 2
        + S1**2 / n
        - S1**2 / (1 + n)
        + S2 / n
        - S2 / (1 + n)
    )


@nb.njit(cache=True)
def lm13m1(n, S1, S2, S3):
    r"""Mellin transform of :math:`(1-x)\ln^3(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[(1-x)\ln^3(1-x)](N)`

    """
    return (
        6 / (1 + n) ** 4
        + (6 * S1) / (1 + n) ** 3
        + (3 * S1**2) / (1 + n) ** 2
        - S1**3 / n
        + S1**3 / (1 + n)
        + (3 * S2) / (1 + n) ** 2
        - (3 * S1 * S2) / n
        + (3 * S1 * S2) / (1 + n)
        - (2 * (6 * (2 * S3 - 2 * zeta3) + zeta3)) / n
        + (2 * (6 * (2 * S3 - 2 * zeta3) + zeta3)) / (1 + n)
    )


@nb.njit(cache=True)
def lm11(n, S1):
    r"""Mellin transform of :math:`\ln(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln(1-x)](N)`

    """
    return -S1 / n


@nb.njit(cache=True)
def lm13(n, S1, S2, S3):
    r"""Mellin transform of :math:`\ln^3(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^3(1-x)](N)`

    """
    return -((S1**3 + 3 * S1 * S2 + 2 * S3) / n)


@nb.njit(cache=True)
def lm14(n, S1, S2, S3, S4):
    r"""Mellin transform of :math:`\ln^4(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`
    S4 : complex
        Harmonic sum :math:`S_{4}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^4(1-x)](N)`

    """
    return (S1**4 + 6 * S1**2 * S2 + 3 * S2**2 + 8 * S1 * S3 + 6 * S4) / n


@nb.njit(cache=True)
def lm15(n, S1, S2, S3, S4, S5):
    r"""Mellin transform of :math:`\ln^5(1-x)`.

    Parameters
    ----------
    n : complex
        Mellin moment
    S1 : complex
        Harmonic sum :math:`S_{1}(N)`
    S2 : complex
        Harmonic sum :math:`S_{2}(N)`
    S3 : complex
        Harmonic sum :math:`S_{3}(N)`
    S4 : complex
        Harmonic sum :math:`S_{4}(N)`
    S5 : complex
        Harmonic sum :math:`S_{5}(N)`

    Returns
    -------
    complex
        :math:`\mathcal{M}[\ln^5(1-x)](N)`

    """
    return (
        -(
            S1**5
            + 10 * S1**3 * S2
            + 20 * S1**2 * S3
            + 15 * S1 * (S2**2 + 2 * S4)
            + 4 * (5 * S2 * S3 + 6 * S5)
        )
        / n
    )



# -*- coding: utf-8 -*-
r"""This subpackage contains some Mellin transformations for 5th order harmonics sum.

Notation refers to:

    - :cite:`Blumlein:2009ta`. Johannes Blumlein. Structural Relations of
      Harmonic Sums and Mellin Transforms up to Weight w = 5. Comput. Phys.
      Commun., 180:2218-2249, 2009. arXiv:0901.3106,
      doi:10.1016/j.cpc.2009.07.004.

Mellin transform is defined with the convention x^(N).
F19, F20, F21 are not present explicitly in the paper
"""  # pylint: disable=line-too-long

from .f9 import F9
from .f11 import F11
from .f13 import F13
from .f14_f12 import F14F12
from .f16 import F16
from .f17 import F17
from .f18 import F18
from .f19 import F19
from .f20 import F20
from .f21 import F21



# -*- coding: utf-8 -*-
import copy

import yaml
from banana.data import sql

from ekomark.data import operators


def gen_op_card(Q2grid, update=None, name=None):
    """
    Generates an operator card with some mandatory user choice
    (in this case only the Q2 grid) and some default values which
    can be changed by the update input dict

    Parameters
    ----------
        Q2grid : list(float)
            grid for Q2
        update : dict
            dictionary of info to update in op. card
        name : str
            name of exported op.card (if name not None)
    Returns
    -------
        : dict
            operator card
    """
    # Constructing the dictionary with some default value
    def_op = copy.deepcopy(operators.default_card)
    # Adding the mandatory inputs
    def_op["Q2grid"] = Q2grid
    if isinstance(update, dict):
        for k in update.keys():
            if k not in def_op.keys():
                raise ValueError("Provided key not in operators card")
        def_op.update(update)
    serialized = sql.serialize(def_op)
    def_op["hash"] = (sql.add_hash(serialized))[-1]
    if name is not None:
        export_op_card(name, def_op)
    return def_op


def export_op_card(name, op):
    """
    Export the operators card in the current directory

    Parameters
    ----------
        name : str
            name of the op. card to export

        op : dict
            op card
    """
    target = f"{name}.yaml"
    with open(target, "w", encoding="utf-8") as out:
        yaml.safe_dump(op, out)


def import_op_card(path):
    """
    Import the operators card specified by path

    Parameters
    ----------
        path : str
            path to op. card in yaml format

    Returns
    -------
        : dict
            op card
    """
    with open(path, "r", encoding="utf-8") as o:
        op = yaml.safe_load(o)
    return op



# -*- coding: utf-8 -*-
import copy
import operator
from numbers import Number

import numpy as np


class OpMember:
    """
    A single operator for a specific element in evolution basis.

    This class provide some basic mathematical operations such as products.
    It can also be applied to a pdf vector via :meth:`apply_pdf`.
    This class will never be exposed to the outside, but will be an internal member
    of the :class:`Operator` and :class:`PhysicalOperator` instances.

    Parameters
    ----------
        value : np.array
            operator matrix
        error : np.array
            operator error matrix
    """

    def __init__(self, value, error):
        self.value = np.array(value)
        self.error = np.array(error)

    def copy(self):
        return self.__class__(self.value.copy(), self.error.copy())

    @classmethod
    def id_like(cls, other):
        """
        Creates an identity operator.

        Parameters
        ----------
            other : OpMember
                reference member

        Returns
        -------
            cls :
                1
        """
        len_xgrid = other.value.shape[0]
        return cls(np.eye(len_xgrid), np.zeros((len_xgrid, len_xgrid)))

    def __matmul__(self, operator_member):
        rval = operator_member.value
        rerror = operator_member.error
        lval = self.value
        ler = self.error
        new_val = np.matmul(lval, rval)
        # TODO check error propagation
        new_err = np.abs(np.matmul(lval, rerror)) + np.abs(np.matmul(ler, rval))
        return self.__class__(new_val, new_err)

    def __mul__(self, other):
        if not isinstance(other, Number):
            raise NotImplementedError(f"Can't multiply OpMember and {type(other)}")
        return self.__class__(other * self.value, other * self.error)

    def __add__(self, operator_member):
        if isinstance(operator_member, Number):
            # we only allow the integer 0 as alias for the true zero operator
            if operator_member != 0:
                raise ValueError(
                    "The only number we can sum to is 0 (as alias for the zero operator)"
                )
            rval = operator_member
            rerror = 0.0
        elif isinstance(operator_member, self.__class__):
            rval = operator_member.value
            rerror = operator_member.error
        else:
            raise NotImplementedError(f"Can't sum OpMember and {type(operator_member)}")
        new_val = self.value + rval
        new_err = self.error + rerror
        return self.__class__(new_val, new_err)

    def __neg__(self):
        return self.__class__(-self.value.copy(), self.error.copy())

    def __eq__(self, operator_member):
        return np.allclose(self.value, operator_member.value)

    def __sub__(self, operator_member):
        return self.__add__(-operator_member)

    def __radd__(self, operator_member):
        return self.__add__(operator_member)

    def __rsub__(self, operator_member):
        return self.__radd__(-operator_member)

    def __rmul__(self, other):
        return self.__mul__(other)


class MemberName:
    """
    Operator member name in operator evolution space

    Parameters
    ----------
        name : str
            operator name
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(str(self))

    def _split_name(self):
        """Splits the name according to target.input"""
        # we need to do this late, as in raw mode the name to not follow this principle
        name_spl = self.name.split(".")
        if len(name_spl) != 2:
            raise ValueError("The operator name has no valid format: target.input")
        for k in [0, 1]:
            name_spl[k] = name_spl[k].strip()
            if len(name_spl[k]) <= 0:
                raise ValueError("The operator name has no valid format: target.input")
        return name_spl

    @property
    def target(self):
        """Returns target flavor name (given by the first part of the name)"""
        return self._split_name()[0]

    @property
    def input(self):
        """Returns input flavor name (given by the second part of the name)"""
        return self._split_name()[1]


class OperatorBase:
    """
    Abstract base class to hold a dictionary of interpolation matrices.

    Parameters
    ----------
        op_members : dict
            mapping of :class:`MemberName` onto :class:`OpMember`
        q2_final : float
            final scale
    """

    def __init__(self, op_members, q2_final):
        self.op_members = op_members
        self.q2_final = q2_final

    def __getitem__(self, key):
        if not isinstance(key, MemberName):
            key = MemberName(key)
        return self.op_members[key]

    @classmethod
    def promote_names(cls, op_members, q2_final):
        """
        Promote string keys to MemberName.

        Parameters
        ----------
            op_members : dict
                mapping of :data:`str` onto :class:`OpMember`
            q2_final : float
                final scale
        """
        # map key to MemberName
        opms = {}
        for k, v in op_members.items():
            opms[MemberName(k)] = copy.copy(v)
        return cls(opms, q2_final)

    def __matmul__(self, other):
        """
        Multiply ``other`` to self.

        Parameters
        ----------
            other : OperatorBase
                second factor with a lower initial scale

        Returns
        -------
            p : PhysicalOperator
                self @ other
        """
        if not isinstance(other, OperatorBase):
            raise ValueError("Can only multiply with another OperatorBase")
        # if a ScalarOperator is multiplied by an OperatorBase, the result is an OperatorBase
        # only the result of ScalarOperator @ ScalarOperator is another ScalarOperator
        cls = self.__class__
        if isinstance(self, ScalarOperator):
            cls = other.__class__
        return cls(
            self.operator_multiply(self, other, self.operation(other)), self.q2_final
        )

    def operation(self, other):
        """
        Choose mathematical operation by rank

        Parameters
        ----------
            other : OperatorBase
                operand

        Returns
        -------
            callable :
                operation to perform (np.matmul or np.multiply)
        """
        if isinstance(self, ScalarOperator) or isinstance(other, ScalarOperator):
            return operator.mul
        return operator.matmul

    @staticmethod
    def operator_multiply(left, right, operation):
        """
        Multiply two operators.

        Parameters
        ----------
            left : OperatorBase
                left operand
            right : OperatorBase
                right operand
            operation : callable
                operation to perform (np.matmul or np.multiply)

        Returns
        -------
            dict :
                new operator members dictionary
        """
        # prepare paths
        new_oms = {}
        for l_key, l_op in left.op_members.items():
            for r_key, r_op in right.op_members.items():
                # ops match?
                if l_key.input != r_key.target:
                    continue
                new_key = MemberName(l_key.target + "." + r_key.input)
                # new?
                if new_key not in new_oms:
                    new_oms[new_key] = operation(l_op, r_op)
                else:  # add element
                    new_oms[new_key] += operation(l_op, r_op)
        return new_oms


class ScalarOperator(OperatorBase):
    pass



# -*- coding: utf-8 -*-
"""Weight 1 harmonic sums."""

import numba as nb
import numpy as np

from .polygamma import cern_polygamma


@nb.njit(cache=True)
def S1(N):
    r"""Computes the harmonic sum :math:`S_1(N)`.

    .. math::
      S_1(N) = \sum\limits_{j=1}^N \frac 1 j = \psi_0(N+1)+\gamma_E

    with :math:`\psi_0(N)` the digamma function and :math:`\gamma_E` the
    Euler-Mascheroni constant.

    Parameters
    ----------
        N : complex
            Mellin moment

    Returns
    -------
        S_1 : complex
            (simple) Harmonic sum :math:`S_1(N)`

    See Also
    --------
        eko.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`
    """
    return cern_polygamma(N + 1.0, 0) + np.euler_gamma


@nb.njit(cache=True)
def Sm1(N, hS1, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-1}(N)`.

    .. math::
      S_{-1}(N) = \sum\limits_{j=1}^N \frac {(-1)^j} j

    Parameters
    ----------
        N : complex
            Mellin moment
        hS1:  complex
            Harmonic sum :math:`S_{1}(N)`
        is_singlet: bool, None
            symmetry factor: True for singlet like quantities (:math:`\eta=(-1)^N = 1`),
            False for non-singlet like quantities (:math:`\eta=(-1)^N=-1`)
    Returns
    -------
        Sm1 : complex
            Harmonic sum :math:`S_{-1}(N)`

    See Also
    --------
        eko.anomalous_dimension.w1.S1 : :math:`S_1(N)`
    """
    if is_singlet is None:
        return (
            (1 - (-1) ** N) / 2 * S1((N - 1) / 2)
            + ((-1) ** N + 1) / 2 * S1(N / 2)
            - hS1
        )
    if is_singlet:
        return S1(N / 2) - hS1
    return S1((N - 1) / 2) - hS1



# -*- coding: utf-8 -*-
"""
This file contains the O(aem2) Altarelli-Parisi splitting kernels.
"""

import numba as nb

from .. import constants
from . import as1aem1


@nb.njit(cache=True)
def gamma_phph(N, nf):
    """Computes the O(aem2) photon-photon singlet anomalous dimension.

    Implements Eq. (68) of :cite:`deFlorian:2016gvk`.

    Parameters
    ----------

    Returns
    -------
        gamma_gg : complex
            O(aem2) photon-photon singlet anomalous dimension
            :math:`\\gamma_{\\gamma \\gamma}^{(0,2)}(N)`

    """

    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return (
        constants.NC
        * (nu * constants.eu2**2 + nd * constants.ed2**2)
        * (as1aem1.gamma_gph(N) / constants.CF / constants.CA + 4)
    )


@nb.njit(cache=True)
def gamma_uph(N, nf, sx):
    """Computes the O(aem2) quark-photon anomalous dimension for up quarks.

    Implements Eq. (55) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
      gamma_uph : complex
        O(aem2) quark-photon anomalous dimension :math:`\\gamma_{u \\gamma}^{(0,2)}(N)`

    """
    return constants.eu2 * as1aem1.gamma_qph(N, nf, sx) / constants.CF


@nb.njit(cache=True)
def gamma_dph(N, nf, sx):
    """Computes the O(aem2) quark-photon anomalous dimension for down quarks.

    Implements Eq. (55) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
      gamma_dph : complex
        O(aem2) quark-photon anomalous dimension :math:`\\gamma_{d \\gamma}^{(0,2)}(N)`

    """
    return constants.ed2 * as1aem1.gamma_qph(N, nf, sx) / constants.CF


@nb.njit(cache=True)
def gamma_phu(N, nf, sx):
    """Computes the O(aem2) photon-quark anomalous dimension for up quarks.

    Implements Eq. (56) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
      gamma_phu : complex
        O(aem2) photon-quark anomalous dimension :math:`\\gamma_{\\gamma u}^{(0,2)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    S1 = sx[0]
    tmp = (-16 * (-16 - 27 * N - 13 * N**2 - 8 * N**3)) / (
        9.0 * (-1 + N) * N * (1 + N) ** 2
    ) - 16 * (2 + 3 * N + 2 * N**2 + N**3) / (
        3.0 * (-1 + N) * N * (1 + N) ** 2
    ) * S1
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    return constants.eu2 * as1aem1.gamma_phq(N, sx) / constants.CF + eSigma2 * tmp


@nb.njit(cache=True)
def gamma_phd(N, nf, sx):
    """Computes the O(aem2) photon-quark anomalous dimension for down quarks.

    Implements Eq. (56) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
      gamma_phd : complex
        O(aem2) photon-quark anomalous dimension :math:`\\gamma_{\\gamma d}^{(0,2)}(N)`

    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    S1 = sx[0]
    tmp = (-16 * (-16 - 27 * N - 13 * N**2 - 8 * N**3)) / (
        9.0 * (-1 + N) * N * (1 + N) ** 2
    ) - 16 * (2 + 3 * N + 2 * N**2 + N**3) / (
        3.0 * (-1 + N) * N * (1 + N) ** 2
    ) * S1
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    return constants.ed2 * as1aem1.gamma_phq(N, sx) / constants.CF + eSigma2 * tmp


@nb.njit(cache=True)
def gamma_nspu(N, nf, sx):
    """Computes the O(aem2) singlet-like non-singlet anomalous dimension for up quarks.

    Implements sum of Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_nspu : complex
            O(aem2) singlet-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,+,u}^{(0,2)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2
        * (-12 + 20 * N + 47 * N**2 + 6 * N**3 + 3 * N**4)
        / (9.0 * N**2 * (1 + N) ** 2)
        - 80 / 9 * S1
        + 16 / 3 * S2
    ) * eSigma2
    return constants.eu2 * as1aem1.gamma_nsp(N, sx) / constants.CF / 2 + tmp


@nb.njit(cache=True)
def gamma_nspd(N, nf, sx):
    """Computes the O(aem2) singlet-like non-singlet anomalous dimension for down quarks.

    Implements sum of Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_nspd : complex
            O(aem2) singlet-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,+,d}^{(0,2)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2
        * (-12 + 20 * N + 47 * N**2 + 6 * N**3 + 3 * N**4)
        / (9.0 * N**2 * (1 + N) ** 2)
        - 80 / 9 * S1
        + 16 / 3 * S2
    ) * eSigma2
    return constants.ed2 * as1aem1.gamma_nsp(N, sx) / constants.CF / 2 + tmp


@nb.njit(cache=True)
def gamma_nsmu(N, nf, sx):
    """Computes the O(aem2) valence-like non-singlet anomalous dimension for up quarks.

    Implements difference between Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=u.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_nsp : complex
            O(aem2) valence-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,-,u}^{(0,2)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2
        * (-12 + 20 * N + 47 * N**2 + 6 * N**3 + 3 * N**4)
        / (9.0 * N**2 * (1 + N) ** 2)
        - 80 / 9 * S1
        + 16 / 3 * S2
    ) * eSigma2
    return constants.eu2 * as1aem1.gamma_nsm(N, sx) / constants.CF / 2 + tmp


@nb.njit(cache=True)
def gamma_nsmd(N, nf, sx):
    """Computes the O(aem2) valence-like non-singlet anomalous dimension for down quarks.

    Implements difference between Eqs. (57-58) of :cite:`deFlorian:2016gvk` for q=d.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors
        sx : np array
            List of harmonic sums

    Returns
    -------
        gamma_nsp : complex
            O(aem2) valence-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,-,d}^{(0,2)}(N)`

    """
    S1 = sx[0]
    S2 = sx[1]
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    eSigma2 = constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    tmp = (
        2
        * (-12 + 20 * N + 47 * N**2 + 6 * N**3 + 3 * N**4)
        / (9.0 * N**2 * (1 + N) ** 2)
        - 80 / 9 * S1
        + 16 / 3 * S2
    ) * eSigma2
    return constants.ed2 * as1aem1.gamma_nsm(N, sx) / constants.CF / 2 + tmp


@nb.njit(cache=True)
def gamma_ps(N, nf):
    """Computes the O(aem2) pure-singlet quark-quark anomalous dimension.

    Implements Eq. (59) of :cite:`deFlorian:2016gvk`.

    Parameters
    ----------
        N : complex
            Mellin moment
        nf : int
            Number of active flavors

    Returns
    -------
        gamma_ps : complex
            O(aem2) pure-singlet quark-quark anomalous dimension
            :math:`\\gamma_{ps}^{(0,2)}(N)`

    """
    result = (
        -4
        * (2 + N * (5 + N))
        * (4 + N * (4 + N * (7 + 5 * N)))
        / ((-1 + N) * N**3 * (1 + N) ** 3 * (2 + N) ** 2)
    )
    return 2 * nf * constants.CA * result



# -*- coding: utf-8 -*-
import copy

import yaml
from banana.data import sql, theories


def gen_theory_card(pto, initial_scale, update=None, name=None):
    """
    Generates a theory card with some mandatory user choice and some
    default values which can be changed by the update input dict

    Parameters
    ----------

        pto : int
            perturbation theory order
        initial_scale: float
            initial scale of evolution
        update : dict
            info to update to default theory card
        name : str
            name of exported theory card (if name not None )

    Returns
    -------

        : dict
        theory card
    """
    # Constructing the dictionary with some default values
    theory = copy.deepcopy(theories.default_card)
    # delete unuseful member
    del theory["FNS"]
    # Adding the mandatory inputs
    theory["PTO"] = pto
    theory["Q0"] = initial_scale
    # Update user choice
    if update is not None:
        for k in update.keys():
            if k not in theory.keys():
                raise ValueError("Provided key not in theory card")
        theory.update(update)
    serialized = sql.serialize(theory)
    theory["hash"] = (sql.add_hash(serialized))[-1]
    if name is not None:
        export_theory_card(name, theory)
    return theory


def export_theory_card(name, theory):
    """
    Export the theory card in the current directory

    Parameters
    ----------
        name : str
            name of the theory card to export

        theory : dict
            theory card
    """
    target = f"{name}.yaml"
    with open(target, "w", encoding="utf-8") as out:
        yaml.safe_dump(theory, out)


def import_theory_card(path):
    """
    Import the theory card specified by path

    Parameters
    ----------
        path : str
            path to theory card in yaml format

    Returns
    -------
        : dict
            theory card
    """
    with open(path, "r", encoding="utf-8") as o:
        theory = yaml.safe_load(o)
    return theory



# -*- coding: utf-8 -*-
r"""
This module contains the scale variation for ``ModSV=exponentiated``.
"""
import numba as nb

from .. import beta


@nb.njit(cache=True)
def gamma_variation(gamma, order, nf, L):
    """
    Adjust the anomalous dimensions with the scale variations.

    Parameters
    ----------
        gamma : numpy.ndarray
            anomalous dimensions
        order : tuple(int,int)
            perturbation order
        nf : int
            number of active flavors
        L : float
            logarithmic ratio of factorization and renormalization scale

    Returns
    -------
        gamma : numpy.ndarray
            adjusted anomalous dimensions
    """
    # since we are modifying *in-place* be carefull, that the order matters!
    # and indeed, we need to adjust the high elements first
    beta0 = beta.beta_qcd((2, 0), nf)
    beta1 = beta.beta_qcd((3, 0), nf)
    if order[0] >= 4:
        gamma[3] -= (
            3 * beta0 * L * gamma[2]
            + (2 * beta1 * L - 3 * beta0**2 * L**2) * gamma[1]
            + (
                beta.beta_qcd((4, 0), nf) * L
                - 5 / 2 * beta1 * beta0 * L**2
                + beta0**3 * L**3
            )
            * gamma[0]
        )
    if order[0] >= 3:
        gamma[2] -= (
            2 * beta0 * gamma[1] * L + (beta1 * L - beta0**2 * L**2) * gamma[0]
        )
    if order[0] >= 2:
        gamma[1] -= beta0 * gamma[0] * L
    return gamma
